import logging
import inspect
import jsonpickle
import re
import simplejson
import threading
from contextlib import contextmanager
from functools import wraps
from pathlib import Path
from typing import (
    Any,
    Tuple,
    Set,
)

class Logger(logging.getLoggerClass()):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._suppressed: Set[Tuple[threading.Thread, int]] = set()
        self._json_pickle_dumps = lambda d: jsonpickle.dumps(d, max_depth=2, unpicklable=False, indent=2)
        self._children: 'list[Logger]' = []
        self.manager.loggerClass = Logger
        self.msg_char_limit = 0
        self._suppressed_lock = threading.Lock()

    def _copy_suppressed(self):
        with self._suppressed_lock:
            copy = self._suppressed.copy()
        return copy

    def _add_suppressed(self, value: Tuple[threading.Thread, int]):
        with self._suppressed_lock:
            self._suppressed.add(value)

    def _remove_suppressed(self, value: Tuple[threading.Thread, int]):
        with self._suppressed_lock:
            self._suppressed.remove(value)

    def _log(self, level, msg, args, exc_info=None, truncate=True, **kwargs) -> None:
        current_thread = threading.current_thread()
        for s_thread, s_level in self._copy_suppressed():
            if s_thread == current_thread and level < s_level:
                return
        if truncate and not exc_info and isinstance(self.msg_char_limit, int) and self.msg_char_limit > 0:
            msg = msg[:self.msg_char_limit]
        return super()._log(level, msg, args=args, exc_info=exc_info, **kwargs)

    def debug(
        self, *args, exc_info=None, stack_info=None, stacklevel=1,
        extra=None, truncate=True, **kwargs: Any
    ) -> None:
        kwargs['truncate'] = truncate
        return super().debug(
            *args, exc_info=exc_info, stack_info=stack_info,
            stacklevel=stacklevel + 2, extra=extra, **kwargs
        )

    def info(
        self, *args, exc_info=None, stack_info=None, stacklevel=1,
        extra=None, truncate=True, **kwargs: Any
    ) -> None:
        kwargs['truncate'] = truncate
        return super().info(
            *args, exc_info=exc_info, stack_info=stack_info,
            stacklevel=stacklevel + 2, extra=extra, **kwargs
        )

    def warning(
        self, *args, exc_info=None, stack_info=None, stacklevel=1,
        extra=None, truncate=True, **kwargs: Any
    ) -> None:
        kwargs['truncate'] = truncate
        return super().warning(
            *args, exc_info=exc_info, stack_info=stack_info,
            stacklevel=stacklevel + 2, extra=extra, **kwargs
        )

    def error(
        self, *args, exc_info=None, stack_info=None, stacklevel=1,
        extra=None, truncate=True, **kwargs: Any
    ) -> None:
        kwargs['truncate'] = truncate
        return super().error(
            *args, exc_info=exc_info, stack_info=stack_info,
            stacklevel=stacklevel + 2, extra=extra, **kwargs
        )

    def critical(
        self, *args, exc_info=None, stack_info=None, stacklevel=1,
        extra=None, truncate=True, **kwargs: Any
    ) -> None:
        kwargs['truncate'] = truncate
        return super().critical(
            *args, exc_info=exc_info, stack_info=stack_info,
            stacklevel=stacklevel + 2, extra=extra, **kwargs
        )

    def exception(
        self, *args, exc_info=True, stack_info=None, stacklevel=1,
        extra=None, truncate=True, **kwargs: Any
    ) -> None:
        kwargs['truncate'] = truncate
        return super().exception(
            *args, exc_info=exc_info, stack_info=stack_info,
            stacklevel=stacklevel + 2, extra=extra, **kwargs
        )

    @contextmanager
    def suppressed(self, level: int, include_child_loggers: bool = True):
        thread = threading.current_thread()
        self._add_suppressed((thread, level))
        if include_child_loggers:
            for cl in self._children:
                cl._add_suppressed((thread, level))
        yield
        try:
            self._remove_suppressed((thread, level))
            if include_child_loggers:
                for cl in self._children:
                    cl._remove_suppressed((thread, level))
        except:
            # Don't let this exception stop the running program.
            pass

    def suppress_function(self, level: int, include_child_loggers: bool = True):
        def dec(f):
            @wraps(f)
            def wrap(*args, **kwargs):
                with self.suppressed(level=level, include_child_loggers=include_child_loggers):
                    return f(*args, **kwargs)

            return wrap

        return dec

    def wrap_class(
        self, level: int, include: str = '', exclude: str = '__.*',
        *logging_args, **logging_kwargs
    ):
        def wrap(cls):
            for attr, fn in inspect.getmembers(cls, inspect.isroutine):
                if callable(getattr(cls, attr)):
                    if include:
                        matches = re.findall(pattern=include, string=fn.__name__, flags=re.IGNORECASE)
                        if fn.__name__ not in matches:
                            continue
                    if exclude:
                        matches = re.findall(pattern=exclude, string=fn.__name__, flags=re.IGNORECASE)
                        if fn.__name__ in matches:
                            continue
                    setattr(
                        cls, attr, self.wrap_func(
                            level=level, _class=cls, *logging_args, **logging_kwargs
                        )(getattr(cls, attr))
                    )
            return cls

        return wrap

    def wrap_func(self, level: int, _class=None, *logging_args, **logging_kwargs):
        def dec(func):
            if func.__name__.startswith('__'):
                return func

            is_staticmethod = isinstance(inspect.getattr_static(_class, func.__name__, None), staticmethod)
            is_classmethod = isinstance(inspect.getattr_static(_class, func.__name__, None), classmethod)

            unwrapped_func = inspect.unwrap(func)
            extra = {
                '_funcName': unwrapped_func.__qualname__,
                '_lineno'  : unwrapped_func.__code__.co_firstlineno,
                '_filename': Path(unwrapped_func.__code__.co_filename).name,
                '_pathname': unwrapped_func.__code__.co_filename,
                '_module'  : inspect.getmodulename(unwrapped_func.__code__.co_filename)
            }

            @wraps(func)
            def wrap(*args, **kwargs):
                in_msg_dict = {
                    'CALLED': func.__qualname__
                }
                if _class and (is_classmethod or is_staticmethod):
                    args = [a for e, a in enumerate(args) if e > 0]
                if args:
                    in_msg_dict['ARGS'] = args
                if kwargs:
                    in_msg_dict['KWARGS'] = kwargs
                self.log(
                    level=level, msg=self._json_pickle_dumps(in_msg_dict), extra=extra,
                    *logging_args, **logging_kwargs
                )
                try:
                    result = func(*args, **kwargs)
                    out_msg_dict = {
                        'RETURNED': func.__qualname__
                    }
                    if result:
                        out_msg_dict['OUTPUT'] = result
                    self.log(
                        level=level, msg=self._json_pickle_dumps(out_msg_dict), extra=extra,
                        *logging_args, **logging_kwargs
                    )
                    return result
                except:
                    self.exception(f'An error occurred in {extra["_funcName"]}.', extra=extra)
                    raise

            return wrap

        return dec

    def getChild(self, suffix: str) -> 'Logger':
        self.manager.loggerClass = Logger
        child = super().getChild(suffix=suffix)
        self._children.append(child)
        return child

def get_logger(name: str = None) -> Logger:
    _original_global_logging_class = logging.getLoggerClass()
    if not isinstance(_original_global_logging_class, type) or not issubclass(
        _original_global_logging_class, logging.Logger
    ):
        _original_global_logging_class = logging.Logger

    _original_root_manager_logging_class = logging.root.manager.loggerClass
    if not isinstance(_original_root_manager_logging_class, type) or not issubclass(
        _original_root_manager_logging_class, logging.Logger
    ):
        _original_root_manager_logging_class = logging.Logger
    # noinspection PyUnresolvedReferences
    logging._acquireLock()
    try:
        logging.setLoggerClass(Logger)
        logging.root.manager.setLoggerClass(Logger)
        new_logger = logging.getLogger(name=name)
        logging.setLoggerClass(_original_global_logging_class)
        logging.root.manager.setLoggerClass(_original_root_manager_logging_class)
        return new_logger
    finally:
        # noinspection PyUnresolvedReferences
        logging._releaseLock()

logger = get_logger('venafi')
api_logger = logger.getChild('api')
features_logger = logger.getChild('features')

json_pickler = jsonpickle
json_pickler.set_encoder_options(simplejson.__name__, sort_keys=True, indent=2)



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class CSPKeyContainerAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "CSP Key Container"
    certificate_vault_id = Attribute('Certificate Vault Id')
    options = Attribute('Options')
    private_key_vault_id = Attribute('Private Key Vault Id')
    private_signing_key_vault_id = Attribute('Private Signing Key Vault Id')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.driver_base import DriverBaseAttributes

class LayoutDriverBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Layout Driver Base"



from __future__ import annotations

import logging
import os
import re
import time
from typing import (
    TYPE_CHECKING,
    Union,
)

from packaging.version import Version

from pyvenafi.logger import (
    features_logger,
    logger,
)
from pyvenafi.tpp.api.websdk.enums.secret_store import Namespaces
from pyvenafi.tpp.api.websdk.models import (
    config,
    identity as ident,
)
from pyvenafi.tpp.features.definitions.exceptions import (
    InvalidResultCode,
    ObjectDoesNotExist,
)

if TYPE_CHECKING:
    from pyvenafi.tpp.api.authenticate import Authenticate

def feature(name: str):
    def decorate(cls):
        if int(os.getenv('PYVENAFI_DOC_IN_PROGRESS', 0)):
            return cls
        setattr(cls, '__feature__', name)
        return features_logger.wrap_class(
            level=logging.DEBUG,
            exclude='_.*'
        )(cls)

    return decorate

class FeatureBase:
    def __init__(self, api: 'Authenticate'):
        self._api = api

    def _config_create(
        self, name: str, parent_folder_dn: str, config_class: str, attributes: dict = None,
        get_if_already_exists: bool = True
    ):
        if attributes:
            attributes = self._name_value_list(attributes=attributes)

        dn = f'{parent_folder_dn}\\{name}'
        response = self._api.websdk.Config.Create.post(
            object_dn=dn, class_name=str(config_class),
            name_attribute_list=attributes or []
        )
        result = response.result
        if result.code != 1:
            if result.code == 401 and get_if_already_exists:
                return self._get_config_object(object_dn=dn)
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

        return response.object

    def _config_delete(self, object_dn, recursive: bool = False):
        result = self._api.websdk.Config.Delete.post(object_dn=object_dn, recursive=recursive).result
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def _get_config_object(
        self, object_dn: str = None, object_guid: str = None,
        raise_error_if_not_exists: bool = True, valid_class_names: list[str] = None
    ):
        if not (object_dn or object_guid):
            raise ValueError(
                'Must supply either an Object DN or Object GUID, but neither was provided.'
            )
        if isinstance(object_dn, config.Object):
            obj = object_dn
        elif isinstance(object_guid, config.Object):
            obj = object_guid
        else:
            response = self._api.websdk.Config.IsValid.post(object_dn=object_dn, object_guid=object_guid)
            if response.result.code == 400:
                if raise_error_if_not_exists:
                    raise ObjectDoesNotExist(f'"{object_dn or object_guid}" does not exist.')
                obj = config.Object()
            else:
                obj = response.object
        if valid_class_names and obj.type_name not in valid_class_names:
            valid_class_names = "\n".join(f"*  {i}" for i in valid_class_names)
            raise TypeError(
                f'"{object_dn or object_guid}" exists, but is not the expected class type.\n'
                f'Got type "{obj.type_name}" instead of one of \n{valid_class_names}.'
            )
        return obj

    def _get_identity_object(
        self, prefixed_name: str = None, prefixed_universal: str = None,
        raise_error_if_not_exists: bool = True
    ):
        if not (prefixed_name or prefixed_universal):
            raise ValueError(
                'Must supply either an prefixed_name or prefixed_universal, but neither was provided.'
            )
        if isinstance(prefixed_name, ident.Identity):
            return prefixed_name
        if isinstance(prefixed_universal, ident.Identity):
            return prefixed_universal

        try:
            response = self._api.websdk.Identity.Validate.post(
                identity=self._identity_dict(prefixed_name=prefixed_name, prefixed_universal=prefixed_universal)
            )
            identity = response.identity if response.api_response.content else None
        except:
            identity = None

        if identity is not None:
            return identity
        elif not raise_error_if_not_exists:
            return ident.Identity()
        else:
            raise ObjectDoesNotExist(f'Could not find identity "{prefixed_name or prefixed_universal}".')

    @staticmethod
    def _identity_dict(prefixed_name: str = None, prefixed_universal: str = None):
        """
        Creates an ID object to write to the Identity APIs.

        Args:
            prefixed_name: The prefixed name of the Identity object.
            prefixed_universal: The prefixed universal GUID of the Identity object.

        Returns:
            {
                'PrefixedUniversal': ``prefixed_universal``,
                'PrefixedName': ``prefixed_name``
            }
        """
        d = {}
        if prefixed_name:
            d.update(
                {
                    'PrefixedName': prefixed_name
                }
            )
        if prefixed_universal:
            d.update(
                {
                    'PrefixedUniversal': prefixed_universal
                }
            )
        return d

    @staticmethod
    def _is_prefixed_universal(identity):
        if isinstance(identity, str):
            if ':' not in identity:
                return False
            prefix, identity = identity.split(':', maxsplit=1)
            regex = '^[{]?[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}[}]?$'
            return bool(re.match(pattern=regex, string=identity))
        return False

    @staticmethod
    def _is_obj_guid(obj: str):
        regex = '^[{]?[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}[}]?$'
        return isinstance(obj, str) and bool(re.match(pattern=regex, string=obj))

    def _get_prefixed_name(self, identity: 'Union[ident.Identity, str, None]'):
        if identity is None:
            return None
        if hasattr(identity, 'prefixed_name'):
            return identity.prefixed_name
        if self._is_prefixed_universal(identity):
            return self._get_identity_object(prefixed_universal=identity).prefixed_name
        return identity

    def _get_prefixed_universal(self, identity: 'Union[ident.Identity, str, None]'):
        if identity is None:
            return None
        if hasattr(identity, 'prefixed_universal'):
            return identity.prefixed_universal
        if self._is_prefixed_universal(identity):
            return identity
        return self._get_identity_object(prefixed_name=identity).prefixed_universal

    def _get_dn(self, obj: 'Union[config.Object, str, None]', parent_dn: str = None):
        if obj is None:
            return None
        if hasattr(obj, 'dn'):
            return obj.dn
        if self._is_obj_guid(obj):
            return self._get_config_object(object_guid=obj).dn
        if parent_dn and not obj.startswith(parent_dn):
            obj = parent_dn.rstrip('\\') + '\\' + obj.rstrip('\\')
        if not obj.startswith(r'\VED'):
            return '\\VED\\' + obj.strip("\\")
        else:
            return obj

    def _get_guid(self, obj: 'Union[config.Object, str, None]', parent_dn: str = None):
        if obj is None:
            return None
        if hasattr(obj, 'guid'):
            return obj.guid
        if self._is_obj_guid(obj):
            return obj
        if parent_dn and not obj.startswith(parent_dn):
            obj = parent_dn.rstrip('\\') + f'\\{obj}'
        if not obj.startswith(r'\VED'):
            obj = '\\VED\\' + obj.strip("\\")
        return self._get_config_object(object_dn=obj).guid

    @staticmethod
    def _log_warning_message(msg: str):
        features_logger.warning(msg, stacklevel=2)

    @staticmethod
    def __no_op(*args, **kwargs):
        pass

    # noinspection ALL
    @staticmethod
    def _name_type_value(name: str, type: str, value):
        return {
            'Name' : str(name),
            'Type' : str(type),
            'Value': str(value)
        }

    @staticmethod
    def _name_value_list(attributes: dict[str, list[str]]):
        nvl = []
        for n, v in attributes.items():
            if v is None:
                continue
            if not isinstance(v, (list, tuple, set)):
                v = str(v)
            else:
                v = list(map(str, v))
            nvl.append(config.NameAttribute(name=n, value=v))
        return nvl

    def _secret_store_delete(self, object_dn: str, namespace: str = Namespaces.config):
        result = self._api.websdk.SecretStore.OwnerDelete.post(namespace=namespace, owner=object_dn).result
        if result.code != 0:
            raise InvalidResultCode(code=result.code, code_description=result.secret_store_result)

    def _is_version_compatible(self, minimum: str = '', maximum: str = ''):
        if minimum and self._api._tpp_version <= Version(minimum):
            return False
        if maximum and self._api._tpp_version >= Version(maximum):
            return False
        return True

    class _Timeout:
        def __init__(self, timeout):
            self.timeout = timeout
            self.max_time = timeout + time.time()
            self._cm = logger.suppressed(999)

        def __enter__(self):
            self._cm.__enter__()
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._cm.__exit__(exc_type, exc_val, exc_tb)
            return

        def is_expired(self, poll: float = 0.5):
            if time.time() >= self.max_time:
                return True
            if poll:
                time.sleep(poll)

        @staticmethod
        def poll(seconds: float):
            time.sleep(seconds)



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.agent_base import AgentBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class AgentAttributes(AgentBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Agent"
    action = Attribute('Action')
    active_directory_dn_host_detail = Attribute('Active Directory DN Host Detail')
    active_directory_domain_host_detail = Attribute('Active Directory Domain Host Detail')
    active_directory_host_details = Attribute('Active Directory Host Details')
    active_directory_query_host_detail = Attribute('Active Directory Query Host Detail')
    active_directory_source_dn_host_detail = Attribute('Active Directory Source DN Host Detail')
    address = Attribute('Address')
    configuration_host_details = Attribute('Configuration Host Details')
    custom_host_detail = Attribute('Custom Host Detail')
    environment_host_details = Attribute('Environment Host Details')
    first_update = Attribute('First Update')
    host_os = Attribute('Host OS')
    hostname = Attribute('Hostname')
    last_update = Attribute('Last Update')
    os_id = Attribute('OS Id')



from __future__ import annotations

import json
import re
import time
from datetime import datetime
from typing import (
    Any,
    Callable,
    get_args,
    get_origin,
    Optional,
    Protocol,
    Type,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

import pydantic.main
from packaging.version import Version
from pydantic import (
    BaseModel,
    create_model,
    Field,
    root_validator,
)
from pydantic.fields import Undefined
from requests import (
    HTTPError,
    Response,
)

from pyvenafi.logger import (
    api_logger,
    json_pickler,
)
from pyvenafi.tpp.dn import DN

if TYPE_CHECKING:
    from pyvenafi.tpp.api import WebSDK
    from pydantic.typing import (
        AbstractSetIntStr,
        MappingIntStrAny,
        NoArgAnyCallable,
    )
    from pyvenafi.tpp.api.session import Session

T = TypeVar('T')

class ApiModelMetaclass(pydantic.main.ModelMetaclass):
    def __new__(mcs, name, bases, namespaces, **kwargs):
        annotations_ = namespaces.get('__annotations__', {})
        # This seemed to have caused some issues, but commenting out seems to fix the problem...
        # for base in bases:
        #     annotations.update(getattr(base, '__annotations__', {}))
        for field in annotations_:
            if not field.startswith('__') and get_origin(annotations_[field]) is not Union:
                annotations_[field] = Union[annotations_[field], Any]
        namespaces['__annotations__'] = annotations_
        return super().__new__(mcs, name, bases, namespaces, **kwargs)

# region Endpoint Definitions
class ApiSource(Protocol):
    _host: str
    _username: str
    _password: str
    _token: str
    _base_url: str
    _app_url: str
    _scheme: str
    _session: 'Session'

    def re_authenticate(self): ...

class ApiEndpoint(object):
    """
    This is the backbone of all API definitions. It performs all requests,
    validations, logging, re-authentication, and holds the raw response. This
    class MUST be inherited by all API definitions.
    """

    api_obj: 'WebSDK'

    def __init__(self, api_obj, url: str):
        """
        Args:
            api_obj: This is passed down from the API type object (eg. WebSDK, etc.) and
                represents that type. This type is REQUIRED because it contains the
                authenticated sessions, base URL, and re-authentication methods. It is
                through these properties this class is able to send and receive requests
                to TPP.
            url: This is the URL extension from the base URL.
        """
        self._api_obj = api_obj
        url = url.strip('/')
        if url.startswith(self._api_obj._base_url):
            self._url = url
        else:
            self._url = f'{self._api_obj._app_url}/{url}'
        self._retries = 3
        self.retry_interval = 0.5

    @property
    def retries(self):
        return self._retries + 1  # The first time isn't a "retry".

    @property
    def _session(self) -> 'Session':
        return self._api_obj._session

    def _should_re_authenticate(self, response: 'Response'):
        return response.status_code == 401 and self._api_obj._token is not None

    @staticmethod
    def _rerun_transaction_required(response: Response):
        """
        Uses a regular expression to search the response text for an indication that the transaction was deadlocked
        and needs to be reran.

        Args:
            response: The raw response object.

        Returns:
            Returns ``True`` if the API key expired. Otherwise ``False``.
        """
        return response.status_code == 500 and bool(
            re.match(
                '.*rerun the transaction.*', response.text,
                flags=re.IGNORECASE
            )
        )

    def _delete(self, params: dict = None):
        """
        Performs a DELETE method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='DELETE', data=params)
        exc = None
        for _ in range(self.retries):
            try:
                response = self._session.delete(url=self._url, params=params)
                self._log_response(response=response)
                if self._should_re_authenticate(response=response):
                    self._re_authenticate()
                    # Trigger the retry.
                    continue
                elif self._rerun_transaction_required(response=response):
                    # Trigger the retry.
                    continue
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            time.sleep(self.retry_interval)
        raise exc

    def _get(self, params: dict = None):
        """
        Performs a GET method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Args:
            params: A dictionary of URL parameters to append to the URL.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='GET', data=params)
        exc = None
        for _ in range(self.retries):
            try:
                response = self._session.get(url=self._url, params=params)
                self._log_response(response=response)
                if self._should_re_authenticate(response=response):
                    self._re_authenticate()
                    # Trigger the retry.
                    continue
                elif self._rerun_transaction_required(response=response):
                    # Trigger the retry.
                    continue
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            time.sleep(self.retry_interval)
        raise exc

    def _post(self, data: Union[list, dict]):
        """
        Performs a POST method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Args:
            data: A dictionary of data to send with the URL.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='POST', data=data)
        exc = None
        for _ in range(self.retries):
            try:
                response = self._session.post(url=self._url, data=data)
                self._log_response(response=response)
                if self._should_re_authenticate(response=response):
                    self._re_authenticate()
                    # Trigger the retry.
                    continue
                elif self._rerun_transaction_required(response=response):
                    # Trigger the retry.
                    continue
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            time.sleep(self.retry_interval)
        raise exc

    def _put(self, data: Union[list, dict]):
        """
        Performs a PUT method request. If the response suggests the API Key is expired, then
        a single attempt is made to re-authenticate to TPP. Otherwise, the raw response is returned.

        Args:
            data: A dictionary of data to send with the URL.

        Returns:
            Returns the raw JSON response.
        """
        self._log_rest_call(method='PUT', data=data)
        exc = None
        for _ in range(self.retries):
            try:
                response = self._session.put(url=self._url, data=data)
                self._log_response(response=response)
                if self._should_re_authenticate(response=response):
                    self._re_authenticate()
                    # Trigger the retry.
                    continue
                elif self._rerun_transaction_required(response=response):
                    # Trigger the retry.
                    continue
                return response
            except (ConnectionResetError, ConnectionError) as e:
                exc = e
            time.sleep(self.retry_interval)
        raise exc

    def _re_authenticate(self):
        """
        The current API token expired and the session needs to be re-authenticated.
        """
        api_logger.debug(
            f'{self._api_obj.__class__.__name__} API authentication token expired. Re-authenticating...',
            stacklevel=2
        )
        self._api_obj.re_authenticate()

    def _log_api_deprecated_warning(self, alternate_api: str = None):
        msg = f'API DEPRECATION WARNING: {self._url} is no longer supported by Venafi.'
        if alternate_api:
            msg += f'\nUse {alternate_api} instead.'
        api_logger.warning(msg, stacklevel=2)

    def _log_rest_call(self, method: str, data: dict = None):
        """
        Logs the URL and any additional data. This enforces consistency in logging across all API calls.
        """
        if data:
            payload = json_pickler.dumps(data)
            api_logger.debug(
                f'{method}\nURL: {self._url}\nBODY: {payload}',
                stacklevel=2
            )
        else:
            api_logger.debug(
                msg=f'{method}\nURL: {self._url}',
                stacklevel=2
            )

    def _log_response(self, response: Response):
        """
        Logs the URL, response code, and the content returned by TPP.
        This enforces consistency in logging across all API calls.
        """
        try:
            pretty_json = json_pickler.dumps(response.json())
        except json.JSONDecodeError:
            pretty_json = response.text or response.reason or 'No Content'
        except:
            pretty_json = 'No Content'

        api_logger.debug(
            msg=f'URL: "{self._url}"\nRESPONSE CODE: {response.status_code}\nCONTENT: {pretty_json}',
            stacklevel=2
        )

    @staticmethod
    def _log_warning_message(msg: str):
        api_logger.warning(msg, stacklevel=2)

    def _is_version_compatible(self, minimum: str = '', maximum: str = ''):
        if minimum and self._api_obj.tpp_version <= Version(minimum):
            return False
        if maximum and self._api_obj.tpp_version >= Version(maximum):
            return False
        return True

class WebSdkEndpoint(ApiEndpoint):
    api_obj: 'WebSDK'

# endregion Endpoint Definitions


# region Model And Field Definitions
def ApiField(
    default: Any = None, *, default_factory: 'Optional[NoArgAnyCallable]' = None, alias: str = None, title: str = None,
    description: str = None, exclude: 'Union[AbstractSetIntStr, MappingIntStrAny, Any]' = None,
    include: 'Union[AbstractSetIntStr, MappingIntStrAny, Any]' = None, const: bool = None, gt: float = None,
    ge: float = None, lt: float = None, le: float = None, multiple_of: float = None, max_digits: int = None,
    decimal_places: int = None, min_items: int = None, max_items: int = None, unique_items: bool = None,
    min_length: int = None, max_length: int = None, allow_mutation: bool = True, regex: str = None,
    discriminator: str = None, repr: bool = True, converter: Callable[[Any], Any] = None, **extra: Any
) -> Any:
    if callable(default_factory):
        default = Undefined
    if converter is not None and callable(converter):
        extra['converter'] = converter
    return Field(
        default=default, default_factory=default_factory, alias=alias, title=title, description=description,
        exclude=exclude, include=include, const=const, gt=gt, ge=ge, lt=lt, le=le, multiple_of=multiple_of,
        max_digits=max_digits, decimal_places=decimal_places, min_items=min_items, max_items=max_items,
        unique_items=unique_items, min_length=min_length, max_length=max_length, allow_mutation=allow_mutation,
        regex=regex, discriminator=discriminator, repr=repr, **extra
    )

# region Output Models
def generate_output(response: Response, output_cls: Type[T], root_field: str = None) -> T:
    """
    Args:
        response: Response instance returned by the ``requests`` call to the server.
        output_cls: Custom APIResponse class.
        root_field: In the case that the returned JSON is an array of objects, then the ``root_field``
            is used to assign that value.

    Returns:
        An instance of ``response_cls``.
    """
    try:
        result = response.json()
    except:
        result = {}
    if not isinstance(result, dict):
        result = {
            str(root_field): result
        } if root_field else {}
    elif root_field:
        result = {
            str(root_field): result,
            **result
        }
    response_inst = output_cls(api_response=response, **result)
    try:
        response.raise_for_status()
        if isinstance(result, dict) and (error_key := getattr(response_inst, 'error', 'Unknown')) in result.keys():
            # There is an error message, so raise an exception.
            raise InvalidResponseError('An error occurred: "%s"' % result[error_key], response=response)
        return response_inst
    except HTTPError as error:
        if isinstance(result, dict) and (error_msg := getattr(response_inst, 'error', None)) is not None:
            raise InvalidResponseError(f'An error occurred: "{error_msg}"', response=response) from error
        raise InvalidResponseError(str(error), response=response)

class ObjectModel(BaseModel, metaclass=ApiModelMetaclass):
    class Config:
        arbitrary_types_allowed = True
        allow_population_by_field_name = True
        json_encoders = {
            DN: lambda x: str(x)
        }

    @root_validator(pre=True, allow_reuse=True)
    def _case_insensitive_validator(cls, values: dict):
        new_values = {}
        lowered_values = {k.lower(): v for k, v in values.items()}
        for fk, fv in cls.__fields__.items():
            if fv.name in values:
                new_value = values[fv.name]
            elif fv.alias.lower() in lowered_values:
                new_value = lowered_values[fv.alias.lower()]
            else:
                continue
            type_args = get_args(fv.type_)
            if datetime in type_args and isinstance(new_value, str) and '\\Date' in new_value:
                new_value = re.sub(r'\D', '', new_value)
            elif DN in type_args:
                new_value = DN(new_value)
            if (converter := fv.field_info.extra.get('converter')) is not None:
                new_value = converter(new_value)
            new_values[fv.alias] = new_value
        return new_values

    def with_extra_properties(self, **kwargs) -> 'ObjectModel':
        d = self.__dict__
        d.update(kwargs)
        return create_model(
            f'New{self.__class__.__name__}',
            __base__=self.__class__,  # noqa
            __module__=self.__module__,
            **d
        )()

class RootOutputModel(ObjectModel):
    api_response: Response = ApiField(alias='api_response', exclude=True)

    class Config(ObjectModel.Config):
        fields = {
            'api_response': {
                'exclude': True
            }
        }

    def assert_valid_response(self):
        """
        Use this method when no response property is available after an API call or to simply
        throw an error if the return code is invalid. This simply asserts that a valid response
        status code was returned by TPP.
        """
        self.api_response.raise_for_status()

    def is_valid_response(self):
        """
        Returns ``True`` when the response is valid, meaning a valid return code was returned by
        TPP, otherwise ``False``.
        """
        try:
            self.api_response.raise_for_status()
            return True
        except:
            return False

class WebSdkOutputModel(RootOutputModel):
    error: str = ApiField(alias='Error')

# endregion Output Models


class InvalidResponseError(Exception):
    def __init__(self, msg: str, response: Response):
        self.response = response
        super().__init__(msg, response)
# endregion Response Definitions



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.discovery_statistics import DiscoveryStatisticsAttributes
from pyvenafi.tpp.attributes.schedule_base import ScheduleBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class AgentDiscoveryBaseAttributes(
    DiscoveryStatisticsAttributes,
    ScheduleBaseAttributes,
    TopAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Agent Discovery Base"
    configuration = Attribute('Configuration')
    discovery_exclusion_dn = Attribute('Discovery Exclusion DN')
    protection_key = Attribute('Protection Key')
    report_dn = Attribute('Report DN')
    status = Attribute('Status')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class ServiceModuleAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Service Module"
    driver_arguments = Attribute('Driver Arguments')
    driver_name = Attribute('Driver Name')
    heartbeat_interval = Attribute('Heartbeat Interval')
    interval = Attribute('Interval')
    started_by = Attribute('Started By')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)

class LegacyKeyBaseAttributes(metaclass=IterableMeta):
    __config_class__ = "Legacy Key Base"
    algorithm = Attribute('Algorithm', min_version='21.4')
    approver = Attribute('Approver', min_version='21.4')
    key_bit_strength = Attribute('Key Bit Strength', min_version='21.4')
    key_vault_id = Attribute('Key Vault Id', min_version='21.4')
    protection_key = Attribute('Protection Key', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class GoogleCloudAppAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Google Cloud App"
    file_validation_disabled = Attribute('File Validation Disabled', min_version='21.4')
    google_project_name = Attribute('Google Project Name', min_version='22.3')
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
    target_proxy_name = Attribute('Target Proxy Name', min_version='21.4')
    target_proxy_type = Attribute('Target Proxy Type', min_version='21.4')
    target_region = Attribute('Target Region', min_version='21.4')
    target_resource = Attribute('Target Resource', min_version='21.4')
    timeout = Attribute('Timeout', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class CodeSigningProjectAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Code Signing Project"
    approval_submission_date = Attribute('Approval Submission Date')
    auditor = Attribute('Auditor')
    certificate_issue_flow_dn = Attribute('Certificate Issue Flow DN')
    certificate_owner = Attribute('Certificate Owner')
    code_signing_application_dn = Attribute('Code Signing Application DN')
    flow_instance_macro = Attribute('Flow Instance Macro')
    key_issue_flow_dn = Attribute('Key Issue Flow DN')
    key_owner = Attribute('Key Owner')
    key_use_approver = Attribute('Key Use Approver')
    key_user = Attribute('Key User')
    owner = Attribute('Owner')
    project_delete_flow_dn = Attribute('Project Delete Flow DN')
    purpose = Attribute('Purpose')
    risk_level = Attribute('Risk Level')
    status = Attribute('Status')
    status_message = Attribute('Status Message')



from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import ssh_certificates

class _SSHCertificates(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/SSHCertificates')
        self.CAKeyPair = self._CAKeyPair(api_obj=self._api_obj, url=f'{self._url}/CAKeyPair')
        self.Request = self._Request(api_obj=self._api_obj, url=f'{self._url}/Request')
        self.Retrieve = self._Retrieve(api_obj=self._api_obj, url=f'{self._url}/Retrieve')
        self.Template = self._Template(api_obj=self._api_obj, url=f'{self._url}/Template')

    class _CAKeyPair(WebSdkEndpoint):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.Create = self._Create(api_obj=self._api_obj, url=f'{self._url}/Create')

        class _Create(WebSdkEndpoint):
            def post(
                self, name: str, parent_dn: str = None, key_algorithm: str = None,
                key_storage: str = None, private_key_data: str = None,
                private_key_passphrase: str = None
            ):
                body = {
                    'Name'                : name,
                    'ParentDN'            : parent_dn,
                    'KeyAlgorithm'        : key_algorithm,
                    'KeyStorage'          : key_storage,
                    'PrivateKeyData'      : private_key_data,
                    'PrivateKeyPassphrase': private_key_passphrase
                }

                class Output(WebSdkOutputModel):
                    created_on: datetime = ApiField(alias='CreatedOn')
                    dn: str = ApiField(alias='DN')
                    fingerprint_sha_256: str = ApiField(alias='FingerprintSHA256')
                    guid: str = ApiField(alias='Guid')
                    key_algorithm: str = ApiField(alias='KeyAlgorithm')
                    key_storage: str = ApiField(alias='KeyStorage')
                    name: str = ApiField(alias='Name')
                    processing_details: ssh_certificates.ProcessingDetails = ApiField(alias='ProcessingDetails')
                    public_key_data: str = ApiField(alias='PublicKeyData')
                    response: ssh_certificates.Output = ApiField(alias='Response')

                return generate_output(output_cls=Output, response=self._post(data=body))

    class _Request(WebSdkEndpoint):
        def post(
            self,
            ca_dn: str,
            key_id: str,
            destination_address: str = None,
            extensions: str = None,
            force_command: str = None,
            object_name: str = None,
            origin: str = None,
            policy_dn: str = None,
            principals: list[str] = None,
            public_key_data: str = None,
            source_addresses: list[str] = None,
            validity_period: str = None,
            include_certificate_details: bool = False,
            include_private_key_data: bool = False,
            private_key_passphrase: str = None,
            processing_timeout: int = None
        ):
            body = {
                'CADN'                     : ca_dn,
                'KeyId'                    : key_id,
                'DestinationAddress'       : destination_address,
                'Extensions'               : extensions,
                'ForceCommand'             : force_command,
                'ObjectName'               : object_name,
                'Origin'                   : origin,
                'PolicyDN'                 : policy_dn,
                'Principals'               : principals,
                'PublicKeyData'            : public_key_data,
                'SourceAddresses'          : source_addresses,
                'ValidityPeriod'           : validity_period,
                'IncludeCertificateDetails': include_certificate_details,
                'IncludePrivateKeyData'    : include_private_key_data,
                'PrivateKeyPassphrase'     : private_key_passphrase,
                'ProcessingTimeout'        : processing_timeout
            }

            class Output(WebSdkOutputModel):
                dn: str = ApiField(alias='Dn')
                guid: str = ApiField(alias='Guid')
                processing_details: ssh_certificates.ProcessingDetails = ApiField(alias='ProcessingDetails')
                response: ssh_certificates.Output = ApiField(alias='Response')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _Retrieve(WebSdkEndpoint):
        def post(
            self, dn: str = None, guid: str = None, include_certificate_details: bool = None,
            include_private_key_data: bool = None, private_key_passphrase: str = None
        ):
            body = {
                'DN'                       : dn,
                'Guid'                     : guid,
                'IncludeCertificateDetails': include_certificate_details,
                'IncludePrivateKeyData'    : include_private_key_data,
                'PrivateKeyPassphrase'     : private_key_passphrase
            }

            class Output(WebSdkOutputModel):
                ca_dn: str = ApiField(alias='CADN')
                ca_guid: str = ApiField(alias='CAGuid')
                certificate_data: str = ApiField(alias='CertificateData')
                certificate_details: ssh_certificates.CertificateDetails = ApiField(alias='CertificateDetails')
                dn: str = ApiField(alias='DN')
                guid: str = ApiField(alias='Guid')
                key_id: str = ApiField(alias='KeyID')
                private_key_data: str = ApiField(alias='PrivateKeyData')
                processing_details: ssh_certificates.ProcessingDetails = ApiField(alias='ProcessingDetails')
                public_key_data: str = ApiField(alias='PublicKeyData')
                request_details: ssh_certificates.RequestDetails = ApiField(alias='RequestDetails')
                response: ssh_certificates.Output = ApiField(alias='Response')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _Template(WebSdkEndpoint):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.Available = self._Available(api_obj=self._api_obj, url=f'{self._url}/Available')
            self.Retrieve = self._Retrieve(api_obj=self._api_obj, url=f'{self._url}/Retrieve')

        class _Available(WebSdkEndpoint):
            def get(self):
                class Output(BaseModel):
                    __root__: list[ssh_certificates.AvailableTemplate]

                return generate_output(output_cls=Output, response=self._get())

        class _Retrieve(WebSdkEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.PublicKeyData = self._PublicKeyData(api_obj=self._api_obj, url=f'{self._url}/PublicKeyData')

            def post(self, template_dn: str = None, template_guid: str = None, include_ca_keypair_details: bool = None):
                body = {
                    'DN'                     : template_dn,
                    'Guid'                   : template_guid,
                    'IncludeCAKeyPairDetails': include_ca_keypair_details
                }

                class Output(WebSdkOutputModel):
                    access_control: ssh_certificates.AccessControl = ApiField(alias='AccessControl')
                    api_client: ssh_certificates.APIClient = ApiField(alias='APIClient')
                    ca_keypair_guid: str = ApiField(alias='CAKeyPairGuid')
                    ca_keypair_dn: str = ApiField(alias='CAKeyPairDN')
                    ca_keypair: ssh_certificates.CAKeyPair = ApiField(alias='CAKeyPair')
                    certificate: ssh_certificates.Certificate = ApiField(alias='Certificate')
                    contacts: list[str] = ApiField(default_factory=list, alias='Contacts')
                    created_on: datetime = ApiField(alias='CreatedOn')
                    dn: str = ApiField(alias='DN')
                    guid: str = ApiField(alias='Guid')
                    name: str = ApiField(alias='Name')
                    response: ssh_certificates.Output = ApiField(alias='Response')

                return generate_output(output_cls=Output, response=self._post(data=body))

            class _PublicKeyData(WebSdkEndpoint):
                def get(self, template_dn: str = None, template_guid: str = None):
                    params = {
                        'DN'  : template_dn,
                        'Guid': template_guid,
                    }

                    class Output(WebSdkOutputModel):
                        response: ssh_certificates.Output = ApiField(alias='Response')

                    return generate_output(output_cls=Output, response=self._get(params=params))



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes

class AmazonCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Amazon CA"
    access_key_id = Attribute('Access Key ID', min_version='21.4')
    amazon_private_ca = Attribute('Amazon Private CA', min_version='21.4')
    aws_credential_dn = Attribute('Aws Credential DN', min_version='21.4')
    domain_control_validation = Attribute('Domain Control Validation', min_version='21.4')
    interval = Attribute('Interval', min_version='21.4')
    keys_generated_by_the_ca = Attribute('Keys Generated By The CA', min_version='21.4')
    region_code = Attribute('Region Code', min_version='21.4')
    retrieval_period = Attribute('Retrieval Period', min_version='21.4')
    secret_access_key = Attribute('Secret Access Key', min_version='21.4')
    signature_algorithm = Attribute('Signature Algorithm', min_version='21.4')
    use_private_ca = Attribute('Use Private CA', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)

class _Platform(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Platform')
        self.Delete = self._Delete(api_obj=self._api_obj, url=f'{self._url}/Delete')

    class _Delete(WebSdkEndpoint):
        def post(self, platform: str):
            body = {
                'Platform': platform
            }

            class Output(WebSdkOutputModel):
                success: bool = ApiField(alias='Success')

            return generate_output(response=self._post(data=body), output_cls=Output)



from __future__ import annotations

from datetime import datetime

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class LogEvent(ObjectModel):
    client_timestamp: datetime = ApiField(alias='ClientTimestamp')
    component: str = ApiField(alias='Component')
    component_id: int = ApiField(alias='ComponentId')
    component_subsystem: str = ApiField(alias='ComponentSubsystem')
    data: str = ApiField(alias='Data')
    grouping: int = ApiField(alias='Grouping')
    id: int = ApiField(alias='Id')
    name: str = ApiField(alias='Name')
    server_timestamp: datetime = ApiField(alias='ServerTimestamp')
    severity: str = ApiField(alias='Severity')
    source_ip: str = ApiField(alias='SourceIp')
    text1: str = ApiField(alias='Text1')
    text2: str = ApiField(alias='Text2')
    value1: int = ApiField(alias='Value1')
    value2: int = ApiField(alias='Value2')

class LogEventApplicationDefinition(ObjectModel):
    application_name: str = ApiField(alias='ApplicationName')
    id: int = ApiField(alias='Id')

class LogEventDefinition(ObjectModel):
    data_format: str = ApiField(alias='DataFormat')
    data_title: str = ApiField(alias='DataTitle')
    data_type: str = ApiField(alias='DataType')
    description: str = ApiField(alias='Description')
    grouping_title: str = ApiField(alias='GroupingTitle')
    grouping_type: str = ApiField(alias='GroupingType')
    id: int = ApiField(alias='Id')
    text1_title: str = ApiField(alias='Text1Title')
    text2_title: str = ApiField(alias='Text2Title')
    value1_title: str = ApiField(alias='Value1Title')
    value1_type: str = ApiField(alias='Value1Type')
    value2_title: str = ApiField(alias='Value2Title')
    value2_type: str = ApiField(alias='Value2Type')

class Field(ObjectModel):
    name: str = ApiField(alias='Name')
    label: str = ApiField(alias='Label')
    required: bool = ApiField(alias='Required')

class Script(ObjectModel):
    name: str = ApiField(alias='Name')
    hash: str = ApiField(alias='Hash')
    fields: list[Field] = ApiField(alias='Fields')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.top import TopAttributes

class LogNotificationContainerAttributes(TopAttributes, metaclass=IterableMeta):
    __config_class__ = "Log Notification Container"
    description = Attribute('Description')



from pyvenafi.cloud.api.models import (
    account_service,
    activitylog_service,
    caoperations_service,
    connectors_service,
    edgemanagement_service,
    integrations_service,
    outagedetection_service,
    provisioning_service,
)



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import bus_status

class _BusStatus(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/BusStatus')
        self.GetStatus = self._GetStatus(api_obj=self._api_obj, url=f'{self._url}/GetStatus')
        self.GetMeshDnsName = self._GetMeshDnsName(api_obj=self._api_obj, url=f'{self._url}/GetMeshDnsName')
        self.SetMeshDnsName = self._SetMeshDnsName(api_obj=self._api_obj, url=f'{self._url}/SetMeshDnsName')

    class _GetStatus(WebSdkEndpoint):
        def post(self):
            class Output(WebSdkOutputModel):
                success: bool = ApiField(alias='Status')
                status: bus_status.Status = ApiField(alias='Status')

            return generate_output(output_cls=Output, response=self._post(data={}))

    class _GetMeshDnsName(WebSdkEndpoint):
        def post(self, engine: str):
            body = {
                "Engine": engine,
            }

            class Output(WebSdkOutputModel):
                dns_name: str = ApiField(alias='DnsName')
                mesh_dns_name: str = ApiField(alias='MeshDnsName')
                success: bool = ApiField(alias='Success')
                status: bus_status.Status = ApiField(alias='Status')

            return generate_output(output_cls=Output, response=self._post(data=body))

    class _SetMeshDnsName(WebSdkEndpoint):
        def post(self, engine: str, mesh_dns_name: str):
            body = {
                "Engined"    : engine,
                "MeshDnsName": mesh_dns_name,
            }

            class Output(WebSdkOutputModel):
                success: bool = ApiField(alias='Success')
                status: bus_status.Status = ApiField(alias='Status')

            return generate_output(output_cls=Output, response=self._post(data=body))



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.person import PersonAttributes

class UserAttributes(PersonAttributes, metaclass=IterableMeta):
    __config_class__ = "User"
    creation_date = Attribute('Creation Date', min_version='21.4')
    password = Attribute('Password', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.branch_base import BranchBaseAttributes

class EngineRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Engine Root"
    acme_certificates_folder = Attribute('ACME Certificates Folder', min_version='21.4')
    acme_enabled = Attribute('ACME Enabled', min_version='21.4')
    adaptable_script_max_file_size = Attribute('Adaptable Script Max File Size', min_version='21.4')
    advanced_key_protect = Attribute('Advanced Key Protect', min_version='21.4')
    allow_acme_folder_creation = Attribute('Allow ACME Folder Creation', min_version='21.4')
    authentication_scheme = Attribute('Authentication Scheme', min_version='21.4')
    automatically_apply_filters = Attribute('Automatically Apply Filters', min_version='21.4')
    contract_renewal_date = Attribute('Contract Renewal Date', min_version='23.3')
    disabled_features = Attribute('Disabled Features', min_version='21.4')
    enable_acme_cert_history = Attribute('Enable ACME Cert History', min_version='23.1')
    flow_process_expiry = Attribute('Flow Process Expiry', min_version='21.4')
    key_pair_renewal_flow = Attribute('Key Pair Renewal Flow', min_version='21.4')
    known_cloud_service = Attribute('Known Cloud Service', min_version='21.4')
    known_cloud_service_region = Attribute('Known Cloud Service Region', min_version='21.4')
    large_tree_support = Attribute('Large Tree Support', min_version='21.4')
    managed_by_options = Attribute('Managed By Options', min_version='21.4')
    message_bus_credential = Attribute('Message Bus Credential', min_version='23.3')
    message_bus_instance = Attribute('Message Bus Instance', min_version='23.3')
    message_bus_port = Attribute('Message Bus Port', min_version='23.3')
    message_bus_server = Attribute('Message Bus Server', min_version='23.3')
    message_bus_tls = Attribute('Message Bus Tls', min_version='23.3')
    monitored_uri_verification_engine_info = Attribute('Monitored URI Verification Engine Info', min_version='21.4')
    monitored_uri_verification_identifier = Attribute('Monitored URI Verification Identifier', min_version='21.4')
    monitored_uri_verification_last_check = Attribute('Monitored URI Verification Last Check', min_version='21.4')
    monitored_uri_verification_last_notification = Attribute(
        'Monitored URI Verification Last Notification',
        min_version='21.4'
    )
    monitored_uri_verification_uri_identifier = Attribute(
        'Monitored URI Verification URI Identifier',
        min_version='21.4'
    )
    oam_enabled = Attribute('OAM Enabled', min_version='21.4')
    oam_headervar = Attribute('OAM HeaderVar', min_version='21.4')
    oam_logout_url = Attribute('OAM Logout Url', min_version='21.4')
    operating_environment = Attribute('Operating Environment', min_version='21.4')
    options = Attribute('Options', min_version='21.4')
    pendo_events_last_sent = Attribute('Pendo Events Last Sent', min_version='21.4')
    require_password = Attribute('Require Password', min_version='21.4')
    revocation_check_issuer_dn = Attribute('Revocation Check Issuer DN', min_version='21.4')
    revocation_check_issuer_identifier = Attribute('Revocation Check Issuer Identifier', min_version='21.4')
    secret_key_renewal_flow = Attribute('Secret Key Renewal Flow', min_version='21.4')
    tn_aperture_plugin_vault_id = Attribute('TN Aperture Plugin Vault Id', min_version='21.4')
    tn_public_cas = Attribute('TN Public CAs', min_version='21.4')
    tn_widget_vault = Attribute('TN Widget Vault', min_version='21.4')
    teams_enabled = Attribute('Teams Enabled', min_version='21.4')
    teams_options = Attribute('Teams Options', min_version='21.4')
    teams_required = Attribute('Teams Required', min_version='21.4')
    time_stamping_certificate_dn = Attribute('Time Stamping Certificate DN', min_version='21.4')
    time_stamping_proxy_host_urls = Attribute('Time Stamping Proxy Host URLs', min_version='21.4')
    time_stamping_proxy_options = Attribute('Time Stamping Proxy Options', min_version='21.4')
    timeout = Attribute('Timeout', min_version='21.4')
    trusted_websdk_modules = Attribute('Trusted WebSDK Modules', min_version='21.4')
    usage_tracking = Attribute('Usage Tracking', min_version='21.4')
    websdk_options = Attribute('WebSDK Options', min_version='21.4')
    whitelisted_elevation_commands = Attribute('Whitelisted Elevation Commands', min_version='21.4')
    workflow_interval = Attribute('Workflow Interval', min_version='21.4')
    workflow_last_check = Attribute('Workflow Last Check', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.api.websdk.models import (
    certificate,
    client,
    codesign,
    config,
    config_schema,
    credential,
    discovery,
    flow,
    hsm_api,
    identity,
    log,
    metadata,
    oauth,
    permissions,
    pki,
    preferences,
    processing_engines,
    recycle_bin,
    resultcodes,
    secret_store,
    ssh,
    ssh_certificates,
    stats,
    system_status,
    workflow,
)



from __future__ import annotations

from pyvenafi.tpp.api.websdk.enums.config import (
    ApplicationAttributeValues,
    CertificateAttributeValues,
    CertificateAuthorityAttributeValues,
    ClientGroupsAttributeValues,
    ClientWorkAttributeValues,
    CodeSignAttributeValues,
    CustomFieldAttributeValues,
    DeviceAttributeValues,
    DiscoveryAttributeValues,
    IdentityAttributeValues,
    JumpServerAttributeValues,
    PlacementRulesAttributeValues,
    WorkflowAttributeValues,
)

class AttributeValues:
    Application = ApplicationAttributeValues
    Certificate = CertificateAttributeValues
    CertificateAuthority = CertificateAuthorityAttributeValues
    ClientGroups = ClientGroupsAttributeValues
    ClientWork = ClientWorkAttributeValues
    CodeSign = CodeSignAttributeValues
    CustomField = CustomFieldAttributeValues
    Device = DeviceAttributeValues
    Discovery = DiscoveryAttributeValues
    Identity = IdentityAttributeValues
    JumpServer = JumpServerAttributeValues
    PlacementRules = PlacementRulesAttributeValues
    Workflow = WorkflowAttributeValues



from __future__ import annotations

from datetime import datetime

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)

class Certificate(ObjectModel):
    authentication: str = ApiField(alias='Authentication')
    check_value: str = ApiField(alias='CheckValue')
    created_on: datetime = ApiField(alias='CreatedOn')
    encipherment: bool = ApiField(alias='Encipherment')
    end_date: datetime = ApiField(alias='EndDate')
    environment_type: int = ApiField(alias='EnvironmentType')
    handle: int = ApiField(alias='Handle')
    id: int = ApiField(alias='Id')
    issuer: str = ApiField(alias='Issuer')
    key_context: str = ApiField(alias='KeyContext')
    key_id: str = ApiField(alias='KeyId')
    label: str = ApiField(alias='Label')
    object_type: int = ApiField(alias='ObjectType')
    signing: bool = ApiField(alias='Signing')
    start_date: datetime = ApiField(alias='StartDate')
    subject: str = ApiField(alias='Subject')
    token: bool = ApiField(alias='Token')
    trusted: bool = ApiField(alias='Trusted')
    value: str = ApiField(alias='Value')

class PrivateKey(ObjectModel):
    authentication: bool = ApiField(alias='Authentication')
    created_on: datetime = ApiField(alias='CreatedOn')
    curve: str = ApiField(alias='Curve')
    decrypt: bool = ApiField(alias='Decrypt')
    encipherment: bool = ApiField(alias='Encipherment')
    environment_type: int = ApiField(alias='EnvironmentType')
    exponent: str = ApiField(alias='Exponent')
    handle: int = ApiField(alias='Handle')
    id: str = ApiField(alias='Id')
    key_context: str = ApiField(alias='KeyContext')
    key_id: str = ApiField(alias='KeyId')
    key_type: str = ApiField(alias='KeyType')
    label: str = ApiField(alias='Label')
    modulus: str = ApiField(alias='Modulus')
    object_type: int = ApiField(alias='ObjectType')
    params: str = ApiField(alias='Params')
    private: bool = ApiField(alias='Private')
    sign: bool = ApiField(alias='Sign')
    signing: bool = ApiField(alias='Signing')
    token: bool = ApiField(alias='Token')
    unwrap: bool = ApiField(alias='Unwrap')

class PublicKey(ObjectModel):
    authentication: bool = ApiField(alias='Authentication')
    bits: str = ApiField(alias='Bits')
    created_on: datetime = ApiField(alias='CreatedOn')
    curve: str = ApiField(alias='Curve')
    ec_point: str = ApiField(alias='ECPoint')
    encipherment: bool = ApiField(alias='Encipherment')
    encrypt: str = ApiField(alias='Encrypt')
    environment_type: int = ApiField(alias='EnvironmentType')
    exponent: str = ApiField(alias='Exponent')
    handle: int = ApiField(alias='Handle')
    id: str = ApiField(alias='Id')
    key_context: str = ApiField(alias='KeyContext')
    key_id: str = ApiField(alias='KeyId')
    key_type: str = ApiField(alias='KeyType')
    label: str = ApiField(alias='Label')
    modulus: str = ApiField(alias='Modulus')
    object_type: int = ApiField(alias='ObjectType')
    params: str = ApiField(alias='Params')
    signing: bool = ApiField(alias='Signing')
    token: bool = ApiField(alias='Token')
    verify: str = ApiField(alias='Verify')
    wrap: str = ApiField(alias='Wrap')

class ClientInfo(ObjectModel):
    client_library_name: str = ApiField(alias='ClientLibraryName')
    client_library_version: str = ApiField(alias='ClientLibraryVersion')

class ProcessInfo(ObjectModel):
    command_line: str = ApiField(alias='CommandLine')
    executable: str = ApiField(alias='Executable')
    executable_hash: str = ApiField(alias='ExecutableHash')
    executable_issuer: str = ApiField(alias='ExecutableIssuer')
    executable_signer: str = ApiField(alias='ExecutableSigner')
    executable_size: int = ApiField(alias='ExecutableSize')
    machine: str = ApiField(alias='Machine')
    username: str = ApiField(alias='Username')

class Parameter(ObjectModel):
    hash_alg: str = ApiField(alias='HashAlg')
    iv: str = ApiField(alias='IV')
    mgf: str = ApiField(alias='MGF')
    parameter_type: str = ApiField(alias='ParameterType')
    salt_len: int = ApiField(alias='SaltLen')
    source: int = ApiField(alias='Source')
    source_data: str = ApiField(alias='SourceData')



from __future__ import annotations

from typing import Literal

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)
from pyvenafi.tpp.api.websdk.models.resultcodes import ResultCodes

CredentialType = Literal[
    'Username and Password Credential',
    'Password Credential'
]

class Result(ObjectModel):
    code: int = ApiField()

    @property
    def credential_result(self):
        return ResultCodes.Credential.get(self.code, 'Unknown')

class CredentialInfo(ObjectModel):
    class_name: str = ApiField(alias='ClassName')
    full_name: str = ApiField(alias='FullName')

class NameTypeValue(ObjectModel):
    name: str = ApiField(alias='Name')
    type: str = ApiField(alias='Type')
    value: str = ApiField(alias='Value')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.server_agent_base_device_placement_work import ServerAgentBaseDevicePlacementWorkAttributes

class ServerAgentCertDevicePlacementWorkAttributes(
    ServerAgentBaseDevicePlacementWorkAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Server Agent Cert Device Placement Work"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes

class XolphinCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Xolphin CA"
    brand = Attribute('Brand')
    included_domains = Attribute('Included Domains')
    product = Attribute('Product')
    type = Attribute('Type')



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    CloudApiEndpoint,
    CloudApiOutputModel,
    generate_output,
)
from pyvenafi.cloud.api.models import certificate_service
from typing import Literal

class _certificate_service:
    def __init__(self, api_obj):
        self.outagedetection = self._outagedetection(api_obj=api_obj)
        self.v1 = self._v1(api_obj=api_obj)

    class _outagedetection(CloudApiEndpoint):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='outagedetection')
            self.v1 = self._v1(api_obj=self._api_obj, url=f'{self._url}/v1')

        class _v1(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.certificates = self._certificates(api_obj=self._api_obj, url=f'{self._url}/certificates')
                self.companydomains = self._companydomains(api_obj=self._api_obj, url=f'{self._url}/companydomains')

            class _certificates(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.imports = self._imports(api_obj=self._api_obj, url=f'{self._url}/imports')

                class _imports(CloudApiEndpoint):
                    def __init__(self, *args, **kwargs):
                        super().__init__(*args, **kwargs)

                    def ID(self, id: str):
                        return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                    def post(self, CertificateImportRequest: certificate_service.CertificateImportRequest):
                        data = {**CertificateImportRequest.dict()}

                        class Output(CloudApiOutputModel):
                            CertificateImportResponse: certificate_service.CertificateImportResponse

                        return generate_output(
                            output_cls=Output,
                            response=self._post(data=data),
                            rc_mapping={
                                201: 'CertificateImportResponse'
                            }
                        )

                    class _ID(CloudApiEndpoint):
                        def get(self):
                            class Output(CloudApiOutputModel):
                                CertificateImportStatusDetailResponse: certificate_service.CertificateImportStatusDetailResponse

                            return generate_output(
                                output_cls=Output,
                                response=self._get(params={}),
                                rc_mapping={
                                    200: 'CertificateImportStatusDetailResponse'
                                }
                            )

            class _companydomains(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)

                def ID(self, id: str):
                    return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

                def get(self, status: Literal['AUTHORIZED', 'EXPIRED', 'PENDING']):
                    data = {
                        'status': status,
                    }

                    class Output(CloudApiOutputModel):
                        CompanyDomainResponse: certificate_service.CompanyDomainResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params=data),
                        rc_mapping={
                            200: 'CompanyDomainResponse'
                        }
                    )

                def post(self, CompanyDomainRequest: certificate_service.CompanyDomainRequest):
                    data = {**CompanyDomainRequest.dict()}

                    class Output(CloudApiOutputModel):
                        CompanyDomainResponse: certificate_service.CompanyDomainResponse

                    return generate_output(
                        output_cls=Output,
                        response=self._post(data=data),
                        rc_mapping={
                            201: 'CompanyDomainResponse',
                            202: 'CompanyDomainResponse'
                        }
                    )

                class _ID(CloudApiEndpoint):
                    def delete(self, retireCertificates: bool):
                        data = {
                            'retireCertificates': retireCertificates,
                        }

                        class Output(CloudApiOutputModel):
                            pass

                        return generate_output(output_cls=Output, response=self._delete(params=data))

                    def get(self):
                        class Output(CloudApiOutputModel):
                            CompanyDomainInformation: certificate_service.CompanyDomainInformation

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params={}),
                            rc_mapping={
                                200: 'CompanyDomainInformation'
                            }
                        )

    class _v1(CloudApiEndpoint):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='v1')
            self.certificateblocklist = self._certificateblocklist(
                api_obj=self._api_obj,
                url=f'{self._url}/certificateblocklist'
            )
            self.certificates = self._certificates(api_obj=self._api_obj, url=f'{self._url}/certificates')
            self.companyproperties = self._companyproperties(
                api_obj=self._api_obj,
                url=f'{self._url}/companyproperties'
            )
            self.trustedcacertificates = self._trustedcacertificates(
                api_obj=self._api_obj,
                url=f'{self._url}/trustedcacertificates'
            )

        class _certificateblocklist(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def FINGERPRINT(self, fingerprint: str):
                return self._FINGERPRINT(api_obj=self._api_obj, url=f'{self._url}/{fingerprint}')

            def delete(self):
                class Output(CloudApiOutputModel):
                    pass

                return generate_output(output_cls=Output, response=self._delete(params={}))

            def get(self, limit: int, page: str):
                data = {
                    'limit': limit,
                    'page' : page,
                }

                class Output(CloudApiOutputModel):
                    CertificateBlockListPageResponse: certificate_service.CertificateBlockListPageResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params=data),
                    rc_mapping={
                        200: 'CertificateBlockListPageResponse'
                    }
                )

            def post(self, CertificateBlockListRequest: certificate_service.CertificateBlockListRequest):
                data = {**CertificateBlockListRequest.dict()}

                class Output(CloudApiOutputModel):
                    CertificateBlockListResponse: certificate_service.CertificateBlockListResponse

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        200: 'CertificateBlockListResponse',
                        201: 'CertificateBlockListResponse'
                    }
                )

            class _FINGERPRINT(CloudApiEndpoint):
                def delete(self):
                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._delete(params={}))

                def get(self):
                    class Output(CloudApiOutputModel):
                        str: str

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'str'
                        }
                    )

        class _certificates(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            class _ID(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.encoded = self._encoded(api_obj=self._api_obj, url=f'{self._url}/encoded')

                class _encoded(CloudApiEndpoint):
                    def get(
                        self,
                        chainOrder: Literal['EE_FIRST', 'EE_ONLY', 'ROOT_FIRST'],
                        format: Literal['DER', 'PEM']
                    ):
                        data = {
                            'chainOrder': chainOrder,
                            'format'    : format,
                        }

                        class Output(CloudApiOutputModel):
                            str: str

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params=data),
                            rc_mapping={
                                200: 'str'
                            }
                        )

        class _companyproperties(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def ID(self, id: str):
                return self._ID(api_obj=self._api_obj, url=f'{self._url}/{id}')

            def TYPE(self, type: str):
                return self._TYPE(api_obj=self._api_obj, url=f'{self._url}/{type}')

            def get(self):
                class Output(CloudApiOutputModel):
                    CompanyPropertyResponse: certificate_service.CompanyPropertyResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'CompanyPropertyResponse'
                    }
                )

            def post(self, CompanyPropertyRequest: certificate_service.CompanyPropertyRequest):
                data = {**CompanyPropertyRequest.dict()}

                class Output(CloudApiOutputModel):
                    CompanyPropertyInformation: certificate_service.CompanyPropertyInformation

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        200: 'CompanyPropertyInformation'
                    }
                )

            class _ID(CloudApiEndpoint):
                def put(self, CompanyPropertyRequest: certificate_service.CompanyPropertyRequest):
                    data = {**CompanyPropertyRequest.dict()}

                    class Output(CloudApiOutputModel):
                        CompanyPropertyInformation: certificate_service.CompanyPropertyInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._put(data=data),
                        rc_mapping={
                            200: 'CompanyPropertyInformation'
                        }
                    )

            class _TYPE(CloudApiEndpoint):
                def get(self):
                    class Output(CloudApiOutputModel):
                        CompanyPropertyInformation: certificate_service.CompanyPropertyInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'CompanyPropertyInformation'
                        }
                    )

        class _trustedcacertificates(CloudApiEndpoint):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.deletion = self._deletion(api_obj=self._api_obj, url=f'{self._url}/deletion')

            def FINGERPRINT(self, fingerprint: str):
                return self._FINGERPRINT(api_obj=self._api_obj, url=f'{self._url}/{fingerprint}')

            def get(self):
                class Output(CloudApiOutputModel):
                    TrustedCACertificateResponse: certificate_service.TrustedCACertificateResponse

                return generate_output(
                    output_cls=Output,
                    response=self._get(params={}),
                    rc_mapping={
                        200: 'TrustedCACertificateResponse'
                    }
                )

            def post(self, TrustedCACertificatesRequest: certificate_service.TrustedCACertificatesRequest):
                data = {**TrustedCACertificatesRequest.dict()}

                class Output(CloudApiOutputModel):
                    TrustedCACertificateResponse: certificate_service.TrustedCACertificateResponse

                return generate_output(
                    output_cls=Output,
                    response=self._post(data=data),
                    rc_mapping={
                        200: 'TrustedCACertificateResponse',
                        201: 'TrustedCACertificateResponse'
                    }
                )

            class _FINGERPRINT(CloudApiEndpoint):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.contents = self._contents(api_obj=self._api_obj, url=f'{self._url}/contents')

                def get(self):
                    class Output(CloudApiOutputModel):
                        ExtendedTrustedCACertificateInformation: certificate_service.ExtendedTrustedCACertificateInformation

                    return generate_output(
                        output_cls=Output,
                        response=self._get(params={}),
                        rc_mapping={
                            200: 'ExtendedTrustedCACertificateInformation'
                        }
                    )

                class _contents(CloudApiEndpoint):
                    def get(
                        self,
                        chainOrder: Literal['EE_FIRST', 'EE_ONLY', 'ROOT_FIRST'],
                        format: Literal['DER', 'PEM']
                    ):
                        data = {
                            'chainOrder': chainOrder,
                            'format'    : format,
                        }

                        class Output(CloudApiOutputModel):
                            str: str

                        return generate_output(
                            output_cls=Output,
                            response=self._get(params=data),
                            rc_mapping={
                                200: 'str'
                            }
                        )

            class _deletion(CloudApiEndpoint):
                def post(
                    self,
                    TrustedCACertificateDeletionRequest: certificate_service.TrustedCACertificateDeletionRequest
                ):
                    data = {**TrustedCACertificateDeletionRequest.dict()}

                    class Output(CloudApiOutputModel):
                        pass

                    return generate_output(output_cls=Output, response=self._post(data=data))



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.report_base import ReportBaseAttributes
from pyvenafi.tpp.attributes.report_filter_base import ReportFilterBaseAttributes

class ReportCertificateInventoryAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Report:CertificateInventory"
    grouping = Attribute('Grouping', min_version='21.4')
    options = Attribute('Options', min_version='21.4')
    policydn = Attribute('PolicyDN', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.flow_action_base import FlowActionBaseAttributes

class CodeSigningEnvironmentTemplateUpdateActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Code Signing Environment Template Update Action"



from __future__ import annotations

# noinspection PyUnresolvedReferences
from pyvenafi.tpp.api.websdk.models import (
    certificate as Certificate,
    client as Client,
    config as Config,
    credential as Credential,
    identity as Identity,
    log as Log,
    metadata as CustomFields,
    permissions as Permissions,
    processing_engines as ProcessingEngines,
    secret_store as SecretStore,
    ssh as SSH,
    stats as Stats,
    system_status as SystemStatus,
    workflow as Workflow,
)
from pyvenafi.tpp.features.workflow import RC as ReasonCode  # noqa



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.flow_action_base import FlowActionBaseAttributes

class AdaptableFlowActionBaseAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Adaptable Flow Action Base"
    adaptable_script_max_file_size = Attribute('Adaptable Script Max File Size', min_version='21.4')
    adaptable_script_try_later_max_attempts = Attribute('Adaptable Script Try Later Max Attempts', min_version='21.4')
    adaptable_scripts_directory_name = Attribute('Adaptable Scripts Directory Name', min_version='21.4')
    log_debug = Attribute('Log Debug', min_version='21.4')
    oauth_token_application_id = Attribute('OAuth Token Application Id', min_version='21.4')
    oauth_token_credential = Attribute('OAuth Token Credential', min_version='21.4')
    oauth_token_scope = Attribute('OAuth Token Scope', min_version='21.4')
    powershell_script = Attribute('PowerShell Script', min_version='21.4')
    powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='21.4')
    script_execution_timeout = Attribute('Script Execution Timeout', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class CiscoCSSAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "CiscoCSS"
    certificate_file = Attribute('Certificate File')
    chain_cert = Attribute('Chain Cert')
    ftp_credential = Attribute('FTP Credential')
    ftp_host = Attribute('FTP Host')
    ftp_path = Attribute('FTP Path')
    ftp_port = Attribute('FTP Port')
    ftp_record = Attribute('FTP Record')
    ftp_type = Attribute('FTP Type')
    file_validation_disabled = Attribute('File Validation Disabled')
    network_validation_disabled = Attribute('Network Validation Disabled')
    private_key_file = Attribute('Private Key File')
    reuse_private_key = Attribute('Reuse Private Key')
    ssl_proxy_list = Attribute('SSL Proxy List')
    ssl_server = Attribute('SSL Server')
    server_type = Attribute('Server Type')
    temp_certificate_label = Attribute('Temp Certificate Label')



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import client

class _Client(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Client')
        self.Delete = self._Delete(api_obj=api_obj, url=f'{self._url}/Delete')
        self.Details = self._Details(api_obj=api_obj, url=f'{self._url}/Details')
        self.Work = self._Work(api_obj=api_obj, url=f'{self._url}/Work')

    def get(
        self,
        client_version: str = None,
        client_type: client.ClientType = None,
        host_name: str = None,
        ip_address: str = None,
        last_seen_on: str = None,
        last_seen_on_greater: str = None,
        last_seen_on_less: str = None,
        mac_address: str = None,
        os_name: client.OSNameType = None,
        os_version: str = None,
        region: str = None,
        serial_number: int = None,
        sid: str = None,
        user_name: str = None,
        virtual_machine_id: int = None
    ):
        params = {
            'ClientVersion'    : client_version,
            'client.ClientType': client_type,
            'HostName'         : host_name,
            'IpAddress'        : ip_address,
            'LastSeenOn'       : last_seen_on,
            'LastSeenOnGreater': last_seen_on_greater,
            'LastSeenOnLess'   : last_seen_on_less,
            'MacAddress'       : mac_address,
            'OSName'           : os_name,
            'OSVersion'        : os_version,
            'Region'           : region,
            'SerialNumber'     : serial_number,
            'SID'              : sid,
            'UserName'         : user_name,
            'VirtualMachineId' : virtual_machine_id
        }

        class Output(WebSdkOutputModel):
            clients: list[client.Client] = ApiField(default_factory=list)
            error_description: str = ApiField(alias='error_description')
            error: str = ApiField(alias='error')

        return generate_output(response=self._get(params=params), output_cls=Output, root_field='clients')

    class _Delete(WebSdkEndpoint):
        def post(self, clients: list[client.Client], delete_associated_devices: bool = False):
            body = {
                'Clients'                : clients,
                'DeleteAssociatedDevices': delete_associated_devices
            }

            class Output(WebSdkOutputModel):
                deleted_count: int = ApiField(alias='DeletedCount')
                error_description: str = ApiField(alias='error_description')
                error: str = ApiField(alias='error')

            return generate_output(response=self._post(data=body), output_cls=Output)

    class _Details(WebSdkEndpoint):
        def get(
            self,
            client_version: int = None,
            client_type: str = None,
            host_name: str = None,
            ip_address: str = None,
            last_seen_on: str = None,
            last_seen_on_greater: str = None,
            last_seen_on_less: str = None,
            mac_address: str = None,
            os_name: client.OSNameType = None,
            os_version: str = None,
            region: str = None,
            serial_number: int = None,
            sid: str = None,
            user_name: str = None,
            virtual_machine_id: int = None
        ):
            params = {
                'ClientVersion'    : client_version,
                'ClientType'       : client_type,
                'HostName'         : host_name,
                'IpAddress'        : ip_address,
                'LastSeenOn'       : last_seen_on,
                'LastSeenOnGreater': last_seen_on_greater,
                'LastSeenOnLess'   : last_seen_on_less,
                'MacAddress'       : mac_address,
                'OSName'           : os_name,
                'OSVersion'        : os_version,
                'Region'           : region,
                'SerialNumber'     : serial_number,
                'SID'              : sid,
                'UserName'         : user_name,
                'VirtualMachineId' : virtual_machine_id
            }

            class Output(WebSdkOutputModel):
                details: list[client.ClientDetails] = ApiField(default_factory=list)
                error_description: str = ApiField(alias='error_description')
                error: str = ApiField(alias='error')

            return generate_output(response=self._get(params=params), output_cls=Output, root_field='details')

    class _Work(WebSdkEndpoint):
        def get(self, work_type: client.WorkType = None):
            params = {
                'client.WorkType': work_type
            }

            class Output(WebSdkOutputModel):
                works: list[client.Work] = ApiField(default_factory=list)
                error_description: str = ApiField(alias='error_description')
                error: str = ApiField(alias='error')

            return generate_output(response=self._get(params=params), output_cls=Output, root_field='works')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.ssh_certificate_base import SSHCertificateBaseAttributes

class SSHClientCertificateAttributes(SSHCertificateBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "SSH Client Certificate"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.application_base import ApplicationBaseAttributes

class DataPowerAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "DataPower"
    application_domain = Attribute('Application Domain', min_version='21.4')
    associate_to_cp = Attribute('Associate To CP', min_version='21.4')
    certificate_name = Attribute('Certificate Name', min_version='21.4')
    certificate_only = Attribute('Certificate Only', min_version='21.4')
    chain_cert = Attribute('Chain Cert', min_version='21.4')
    create_cp = Attribute('Create CP', min_version='21.4')
    create_ic = Attribute('Create IC', min_version='21.4')
    create_vc = Attribute('Create VC', min_version='21.4')
    credential_type = Attribute('Credential Type', min_version='21.4')
    crypto_profile = Attribute('Crypto Profile', min_version='21.4')
    ftp_credential = Attribute('FTP Credential', min_version='21.4')
    ftp_host = Attribute('FTP Host', min_version='21.4')
    ftp_path = Attribute('FTP Path', min_version='21.4')
    ftp_port = Attribute('FTP Port', min_version='21.4')
    file_validation_disabled = Attribute('File Validation Disabled', min_version='21.4')
    fips_key = Attribute('Fips Key', min_version='21.4')
    folder = Attribute('Folder', min_version='21.4')
    max_filesize = Attribute('Max Filesize', min_version='21.4')
    network_validation_disabled = Attribute('Network Validation Disabled', min_version='21.4')
    password_alias = Attribute('Password Alias', min_version='24.1')
    private_key_name = Attribute('Private Key Name', min_version='21.4')
    ssh_prompt = Attribute('SSH Prompt', min_version='21.4')
    ssl_profile_type = Attribute('SSL Profile Type', min_version='21.4')
    ssl_proxy_profile = Attribute('SSL Proxy Profile', min_version='21.4')
    temp_certificate_label = Attribute('Temp Certificate Label', min_version='21.4')
    use_basic_provisioning = Attribute('Use Basic Provisioning', min_version='21.4')
    xml_port = Attribute('XML Port', min_version='21.4')



from __future__ import annotations

class KeyContext:
    authentication = 'authentication'
    encryption = 'encryption'
    null = 'null'
    signing = 'signing'

class Mechanism:
    ec_dsa = 4161  # EcDsa
    ec_dsa_sha_1 = 4162  # EcDsaSha1
    ec_dsa_sha_224 = 4163  # EcDsaSha224
    ec_dsa_sha_256 = 4164  # EcDsaSha256
    ec_dsa_sha_384 = 4165  # EcDsaSha384
    ec_dsa_sha_512 = 4166  # EcDsaSha512
    ed_dsa = 4183  # EdDsa
    rsa_pkcs = 1  # RsaPkcs
    rsa_pkcs_oaep = 9  # RsaPkcsOaep
    rsa_pkcs_pss = 13  # RsaPkcsPss
    rsa_pss_sha_1 = 14  # RsaPssSha1
    rsa_pss_sha_224 = 71  # RsaPssSha224
    rsa_pss_sha_256 = 67  # RsaPssSha256
    rsa_pss_sha_384 = 68  # RsaPssSha384
    rsa_pss_sha_512 = 69  # RsaPssSha512
    rsa_sha_1 = 6  # RsaSha1
    rsa_sha_224 = 70  # RsaSha224
    rsa_sha_256 = 64  # RsaSha256
    rsa_sha_384 = 65  # RsaSha384
    rsa_sha_512 = 66  # RsaSha512
    sha_1 = 544  # Sha1
    sha_224 = 597  # Sha224
    sha_256 = 592  # Sha256
    sha_384 = 608  # Sha384
    sha_512 = 624  # Sha512



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)

class _Revoke(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Revoke')
        self._url = self._url.replace('vedsdk', 'vedauth')
        self.Token = self._Token(api_obj=self._api_obj, url=f'{self._url}/Token')

    class _Token(WebSdkEndpoint):
        def get(self):
            response = generate_output(output_cls=WebSdkOutputModel, response=self._get())
            # Set this to None to avoid erroneous re-authentication.
            self._api_obj._token = None
            return response



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    Attribute,
    IterableMeta,
)
from pyvenafi.tpp.attributes.http_ca_base import HTTPCABaseAttributes
from pyvenafi.tpp.attributes.proxy import ProxyAttributes

class VeriSignCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
    __config_class__ = "VeriSign CA"
    additional_field = Attribute('Additional Field')
    admin_isd = Attribute('Admin Isd')
    allow_reissue = Attribute('Allow Reissue', min_version='17.2')
    attempt_renewal = Attribute('Attempt Renewal', min_version='17.2')
    certificate_block = Attribute('Certificate Block')
    certificate_transparency = Attribute('Certificate Transparency', min_version='16.3')
    domain = Attribute('Domain')
    interval = Attribute('Interval')
    jurisdiction_hash = Attribute('Jurisdiction Hash')
    mpki_url = Attribute('Mpki Url')
    organization = Attribute('Organization')
    organizational_unit = Attribute('Organizational Unit')
    renewal_window = Attribute('Renewal Window', min_version='17.2')
    retrieval_period = Attribute('Retrieval Period')
    vice_url = Attribute('Vice Url')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.driver_base import DriverBaseAttributes

class EncryptionDriverAttributes(DriverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Encryption Driver"
    encryption_data = Attribute('Encryption Data')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.monitoring_base import MonitoringBaseAttributes
from pyvenafi.tpp.attributes.top import TopAttributes

class SSHCertificateBaseAttributes(MonitoringBaseAttributes, TopAttributes, metaclass=IterableMeta):
    __config_class__ = "SSH Certificate Base"
    certificate_identifier = Attribute('Certificate Identifier', min_version='21.4')
    certificate_vault_id = Attribute('Certificate Vault Id', min_version='21.4')
    destination_address = Attribute('Destination Address', min_version='21.4')
    error_message = Attribute('Error Message', min_version='21.4')
    extension = Attribute('Extension', min_version='21.4')
    force_command = Attribute('Force Command', min_version='21.4')
    hash_algorithm = Attribute('Hash Algorithm', min_version='21.4')
    in_error = Attribute('In Error', min_version='21.4')
    key_algorithm = Attribute('Key Algorithm', min_version='21.4')
    key_id = Attribute('Key ID', min_version='21.4')
    management_type = Attribute('Management Type', min_version='23.3')
    origin = Attribute('Origin', min_version='21.4')
    originating_ip = Attribute('Originating IP', min_version='21.4')
    principal = Attribute('Principal', min_version='21.4')
    private_key_vault_id = Attribute('Private Key Vault Id', min_version='21.4')
    private_signing_key_vault_id = Attribute('Private Signing Key Vault Id', min_version='21.4')
    public_key_hash = Attribute('Public Key Hash', min_version='21.4')
    public_key_vault_id = Attribute('Public Key Vault Id', min_version='21.4')
    requested_by = Attribute('Requested By', min_version='21.4')
    ssh_ca_template_dn = Attribute('SSH CA Template DN', min_version='21.4')
    ssh_certificate_issuance_flow = Attribute('SSH Certificate Issuance Flow', min_version='21.4')
    serial = Attribute('Serial', min_version='21.4')
    signing_key_fingerprint_sha256 = Attribute('Signing Key Fingerprint SHA256', min_version='21.4')
    source_address = Attribute('Source Address', min_version='21.4')
    stage = Attribute('Stage', min_version='21.4')
    status = Attribute('Status', min_version='21.4')
    valid_from = Attribute('Valid From', min_version='21.4')
    valid_to = Attribute('Valid To', min_version='21.4')
    validity_period = Attribute('Validity Period', min_version='21.4')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.discovery_statistics import DiscoveryStatisticsAttributes
from pyvenafi.tpp.attributes.service_module import ServiceModuleAttributes

class DiscoveryManagerAttributes(DiscoveryStatisticsAttributes, ServiceModuleAttributes, metaclass=IterableMeta):
    __config_class__ = "Discovery Manager"
    connection_timeout = Attribute('Connection Timeout', min_version='21.4')
    delay = Attribute('Delay', min_version='21.4')
    load_percentage = Attribute('Load Percentage', min_version='21.4')
    max_work_units = Attribute('Max Work Units', min_version='21.4')
    maximum_threads = Attribute('Maximum Threads', min_version='21.4')
    minimum_threads = Attribute('Minimum Threads', min_version='21.4')
    placement_disabled = Attribute('Placement Disabled', min_version='21.4')
    timeout = Attribute('Timeout', min_version='21.4')
    window_days_of_week = Attribute('Window Days of Week', min_version='21.4')
    window_end = Attribute('Window End', min_version='21.4')
    window_start = Attribute('Window Start', min_version='21.4')



from __future__ import annotations

from datetime import datetime

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import (
    identity as ident,
    workflow,
)

class _Workflow(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Workflow')
        self.Ticket = self._Ticket(api_obj=api_obj, url=f'{self._url}/Ticket')

    class _Ticket(WebSdkEndpoint):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.Create = self._Create(api_obj=self._api_obj, url=f'{self._url}/Create')
            self.Delete = self._Delete(api_obj=self._api_obj, url=f'{self._url}/Delete')
            self.Details = self._Details(api_obj=self._api_obj, url=f'{self._url}/Details')
            self.Enumerate = self._Enumerate(api_obj=self._api_obj, url=f'{self._url}/Enumerate')
            self.Exists = self._Exists(api_obj=self._api_obj, url=f'{self._url}/Exists')
            self.Status = self._Status(api_obj=self._api_obj, url=f'{self._url}/Status')
            self.UpdateStatus = self._UpdateStatus(api_obj=self._api_obj, url=f'{self._url}/UpdateStatus')

        class _Create(WebSdkEndpoint):
            def post(
                self,
                object_dn: str,
                approvers: list[ident.Identity],
                reason: str,
                workflow_dn: str,
                user_data: str = None
            ):
                body = {
                    'ObjectDN'  : object_dn,
                    'Approvers' : approvers,
                    'Reason'    : reason,
                    'UserData'  : user_data,
                    'WorkflowDN': workflow_dn
                }

                class Output(WebSdkOutputModel):
                    guid: str = ApiField(alias='GUID')
                    result: workflow.Result = ApiField(alias='Result')

                return generate_output(output_cls=Output, response=self._post(data=body))

        class _Delete(WebSdkEndpoint):
            def post(self, guid: str):
                body = {
                    'GUID': guid
                }

                class Output(WebSdkOutputModel):
                    result: workflow.Result = ApiField(alias='Result')

                return generate_output(output_cls=Output, response=self._post(data=body))

        class _Details(WebSdkEndpoint):
            def post(self, guid: str):
                body = {
                    'GUID': guid
                }

                class Output(WebSdkOutputModel):
                    approval_explanation: str = ApiField(alias='ApprovalExplanation')
                    approval_from: str = ApiField(alias='ApprovalFrom')
                    approval_reason: str = ApiField(alias='ApprovalReason')
                    approvers: list[str] = ApiField(alias='Approvers', default_factory=list)
                    blocking: str = ApiField(alias='Blocking')
                    created: datetime = ApiField(alias='Created')
                    issued_due_to: str = ApiField(alias='IssuedDueTo')
                    result: workflow.Result = ApiField(alias='Result')
                    status: str = ApiField(alias='Status')
                    updated: datetime = ApiField(alias='Updated')

                return generate_output(output_cls=Output, response=self._post(data=body))

        class _Enumerate(WebSdkEndpoint):
            def post(self, object_dn: str = None, user_data: str = None):
                body = {
                    'ObjectDN': object_dn,
                    'UserData': user_data
                }

                class Output(WebSdkOutputModel):
                    guids: list[str] = ApiField(default_factory=list, alias='GUIDS')
                    result: workflow.Result = ApiField(alias='Result')

                return generate_output(output_cls=Output, response=self._post(data=body))

        class _Exists(WebSdkEndpoint):
            def post(self, guid: str):
                body = {
                    'GUID': guid
                }

                class Output(WebSdkOutputModel):
                    result: workflow.Result = ApiField(alias='Result')

                return generate_output(output_cls=Output, response=self._post(data=body))

        class _Status(WebSdkEndpoint):
            def post(self, guid: str):
                body = {
                    'GUID': guid
                }

                class Output(WebSdkOutputModel):
                    status: str = ApiField(alias='Status')
                    result: workflow.Result = ApiField(alias='Result')

                return generate_output(output_cls=Output, response=self._post(data=body))

        class _UpdateStatus(WebSdkEndpoint):
            def post(
                self, guid: str, status: str, explanation: str = None, scheduled_start: str = None,
                scheduled_stop: str = None, approvers: list[str] = None, object_dn: str = None,
                reason: str = None, user_data=None
            ):
                body = {
                    'GUID'          : guid,
                    'Status'        : status,
                    'Explanation'   : explanation,
                    'ScheduledStart': scheduled_start,
                    'ScheduledStop' : scheduled_stop,
                    'Approvers'     : approvers,
                    'ObjectDN'      : object_dn,
                    'Reason'        : reason,
                    'UserData'      : user_data
                }

                class Output(WebSdkOutputModel):
                    result: workflow.Result = ApiField(alias='Result')

                return generate_output(output_cls=Output, response=self._post(data=body))



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.credential_base import CredentialBaseAttributes

class AdaptableCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Adaptable Credential"
    credential_connector = Attribute('Credential Connector')
    credential_type = Attribute('Credential Type')
    option_1 = Attribute('Option 1')
    option_2 = Attribute('Option 2')
    password_1 = Attribute('Password 1')
    text_field_1 = Attribute('Text Field 1')
    text_field_2 = Attribute('Text Field 2')
    text_field_3 = Attribute('Text Field 3')
    text_field_4 = Attribute('Text Field 4')
    text_field_5 = Attribute('Text Field 5')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.flow import FlowAttributes

class OAuthDeviceAuthorizationFlowAttributes(FlowAttributes, metaclass=IterableMeta):
    __config_class__ = "OAuth Device Authorization Flow"
    expiration = Attribute('Expiration')
    retry_interval = Attribute('Retry Interval')



from __future__ import annotations

from datetime import datetime
from typing import (
    Literal,
)

from pyvenafi.tpp.api.api_base import (
    ApiField,
    ObjectModel,
)
from pyvenafi.tpp.api.websdk.models.resultcodes import ResultCodes

ClientType = Literal['VenafiAgent', 'AgentJuniorMachine', 'AgentJuniorUser', 'Portal', 'Agentless',
'PreEnrollment', 'iOS', 'Android']
OSNameType = Literal['AIX', 'Android', 'BlackBerry', 'BSD', 'HPux', 'iOS', 'Linux', 'MacOS', 'Other',
'Solaris', 'Unknown', 'Windows', 'zOS']
WorkType = Literal['Client Agent Configuration Work', 'Client Agent Automatic Upgrade Work',
'Client Agent Device Placement Work', 'Certificate Provisioning Work', 'Client Agent SSH Discovery Work',
'Client Agent SSH Provisioning Work', 'Client Agent SSH Key Usage Work', 'Client User Certificate Work',
'Client Certificate Work']

# region Models
class Result(ObjectModel):
    code: int = ApiField()

    @property
    def client_result(self) -> str:
        return ResultCodes.Client.get(self.code, 'Unknown')

class Work(ObjectModel):
    associated_groups: list[str] = ApiField(alias='AssociatedGroups', default_factory=list)
    work_dn: str = ApiField(alias='WorkDn')
    work_name: str = ApiField(alias='WorkName')
    work_type: WorkType = ApiField(alias='WorkType')

class Network(ObjectModel):
    ip_address: str = ApiField(alias='IpAddress')
    mac_address: str = ApiField(alias='MacAddress')

class Client(ObjectModel):
    client_id: int = ApiField(alias='ClientId')
    client_type: ClientType = ApiField(alias='ClientType')
    fqdn: str = ApiField(alias='Fqdn')
    os_name: OSNameType = ApiField(alias='OsName')
    username: str = ApiField(alias='Username')

class ClientDetails(ObjectModel):
    certificate_device: str = ApiField(alias='CertificateDevice')
    client_id: int = ApiField(alias='ClientId')
    client_type: ClientType = ApiField(alias='ClientType')
    client_version: str = ApiField(alias='ClientVersion')
    created_on: datetime = ApiField(alias='CreatedOn')
    dns_name: str = ApiField(alias='DnsName')
    effective_work: list[str] = ApiField(alias='EffectiveWork', default_factory=list)
    fqdn: str = ApiField(alias='Fqdn')
    groups: list[str] = ApiField(alias='Groups', default_factory=list)
    host_domain: str = ApiField(alias='HostDomain')
    hostname: str = ApiField(alias='Hostname')
    last_seen_on: datetime = ApiField(alias='LastSeenOn')
    networks: list[Network] = ApiField(alias='Networks', default_factory=list)
    os_build: str = ApiField(alias='OsBuild')
    os_name: OSNameType = ApiField(alias='OsName')
    os_service_pack: str = ApiField(alias='OsServicePack')
    os_version: str = ApiField(alias='OsVersion')
    region: str = ApiField(alias='Region')
    serial_number: str = ApiField(alias='SerialNumber')
    ssh_device: str = ApiField(alias='SshDevice')
    system_architecture: str = ApiField(alias='SystemArchitecture')
    system_chassis: str = ApiField(alias='SystemChassis')
    system_manufacturer: str = ApiField(alias='SystemManufacturer')
    system_model: str = ApiField(alias='SystemModel')
    trust_level: str = ApiField(alias='TrustLevel')
    username: str = ApiField(alias='Username')
    virtual_machine_id: str = ApiField(alias='VirtualMachineId')



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    ApiField,
    ObjectModel,
)
from datetime import datetime
from typing import (
    Any,
    Dict,
    List,
    Literal,
)
from uuid import UUID

AnyValue = Any

class EnvironmentInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    id: UUID = ApiField(alias='id')
    modificationDate: datetime = ApiField(alias='modificationDate')
    name: str = ApiField(alias='name')
    productEntitlements: List[Literal['ANY', 'CODESIGN', 'DEVOPS', 'MIRA', 'OUTAGE_DETECTION']
    ] = ApiField(alias='productEntitlements', default_factory=list)

class EnvironmentResponse(ObjectModel):
    environments: List[EnvironmentInformation] = ApiField(alias='environments', default_factory=list)

class ErrorInformation(ObjectModel):
    args: List[AnyValue] = ApiField(alias='args', default_factory=list)
    code: int = ApiField(alias='code')
    message: str = ApiField(alias='message')

class ErrorResponse(ObjectModel):
    errors: List[ErrorInformation] = ApiField(alias='errors', default_factory=list)

class IntegrationServiceCreationRequest(ObjectModel):
    edgeInstancesIds: List[UUID] = ApiField(alias='edgeInstancesIds', default_factory=list)
    environmentId: UUID = ApiField(alias='environmentId')
    name: str = ApiField(alias='name')
    scheduleEnabled: bool = ApiField(alias='scheduleEnabled')
    schedulePattern: SchedulerPatternInformation = ApiField(alias='schedulePattern')
    serviceType: Literal['ACME', 'BASIC_DISCOVERY', 'ENHANCED_DISCOVERY', 'EXTERNAL_SCAN', 'INTERNET_DISCOVERY',
    'KEY_GENERATION', 'MSCA', 'SMART_DISCOVERY_EXTERNAL'] = ApiField(alias='serviceType')
    targets: TargetsInformation = ApiField(alias='targets')
    workTypes: List[Literal['DISCOVER', 'MANAGE']] = ApiField(alias='workTypes', default_factory=list)

class IntegrationServiceDetailsResponse(ObjectModel):
    integrationsServices: List[IntegrationServiceInformation] = ApiField(
        alias='integrationsServices',
        default_factory=list
    )
    totalCount: int = ApiField(alias='totalCount')

class IntegrationServiceInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    edgeInstancesIds: List[UUID] = ApiField(alias='edgeInstancesIds', default_factory=list)
    encryptionKeyId: UUID = ApiField(alias='encryptionKeyId')
    environmentId: UUID = ApiField(alias='environmentId')
    id: UUID = ApiField(alias='id')
    name: str = ApiField(alias='name')
    schedulePattern: SchedulerPatternInformation = ApiField(alias='schedulePattern')
    schedulerEnabled: bool = ApiField(alias='schedulerEnabled')
    serviceType: Literal['ACME', 'BASIC_DISCOVERY', 'ENHANCED_DISCOVERY', 'EXTERNAL_SCAN', 'INTERNET_DISCOVERY',
    'KEY_GENERATION', 'MSCA', 'SMART_DISCOVERY_EXTERNAL'] = ApiField(alias='serviceType')
    status: Literal['ACTIVE', 'DISABLED', 'DRAFT', 'ERROR', 'RUNNING', 'WARNING'] = ApiField(alias='status')
    statusMessage: str = ApiField(alias='statusMessage')
    systemGenerated: bool = ApiField(alias='systemGenerated')
    targets: TargetsInformation = ApiField(alias='targets')
    workTypes: List[Literal['DISCOVER', 'MANAGE']] = ApiField(alias='workTypes', default_factory=list)

class IntegrationServiceInitiateScanRequest(ObjectModel):
    serviceId: UUID = ApiField(alias='serviceId')

class IntegrationServiceUpdateRequest(ObjectModel):
    edgeInstancesIds: List[UUID] = ApiField(alias='edgeInstancesIds', default_factory=list)
    name: str = ApiField(alias='name')
    scheduleEnabled: bool = ApiField(alias='scheduleEnabled')
    schedulePattern: SchedulerPatternInformation = ApiField(alias='schedulePattern')
    targets: TargetsInformation = ApiField(alias='targets')
    workTypes: List[Literal['DISCOVER', 'MANAGE']] = ApiField(alias='workTypes', default_factory=list)

class IntegrationsServicesAggregatesResponse(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    integrationsServices: List[Dict[str, Dict[str, Dict[str, int]]]] = ApiField(
        alias='integrationsServices',
        default_factory=list
    )
    totalServicesCount: int = ApiField(alias='totalServicesCount')

class SchedulerPatternInformation(ObjectModel):
    recurrenceType: str = ApiField(alias='recurrenceType')

class TargetsInformation(ObjectModel):
    serviceType: str = ApiField(alias='serviceType')

class WeeklyPatternInformation(SchedulerPatternInformation):
    daysOfWeek: List[Literal['FRIDAY', 'MONDAY', 'SATURDAY', 'SUNDAY', 'THURSDAY',
    'TUESDAY', 'WEDNESDAY']] = ApiField(alias='daysOfWeek', default_factory=list)
    recurrenceTime: datetime = ApiField(alias='recurrenceTime')

class BasicDiscoveryTargetsInformation(TargetsInformation):
    fqdns: List[str] = ApiField(alias='fqdns', default_factory=list)
    ipRanges: List[str] = ApiField(alias='ipRanges', default_factory=list)
    ports: List[str] = ApiField(alias='ports', default_factory=list)
    retireCertificates: bool = ApiField(alias='retireCertificates')

class CronPatternInformation(SchedulerPatternInformation):
    cronExpression: str = ApiField(alias='cronExpression')

class EnhancedDiscoveryTargetsInformation(TargetsInformation):
    fqdns: List[str] = ApiField(alias='fqdns', default_factory=list)
    ipRanges: List[str] = ApiField(alias='ipRanges', default_factory=list)
    ports: List[str] = ApiField(alias='ports', default_factory=list)
    retireCertificates: bool = ApiField(alias='retireCertificates')

class ExternalScanTargetsInformation(TargetsInformation):
    domains: List[str] = ApiField(alias='domains', default_factory=list)
    fqdns: List[str] = ApiField(alias='fqdns', default_factory=list)
    ipRanges: List[str] = ApiField(alias='ipRanges', default_factory=list)
    ports: List[str] = ApiField(alias='ports', default_factory=list)
    retireCertificates: bool = ApiField(alias='retireCertificates')

class InternetDiscoveryTargetsInformation(TargetsInformation):
    domains: List[str] = ApiField(alias='domains', default_factory=list)
    fqdns: List[str] = ApiField(alias='fqdns', default_factory=list)
    ipRanges: List[str] = ApiField(alias='ipRanges', default_factory=list)
    ports: List[str] = ApiField(alias='ports', default_factory=list)
    retireCertificates: bool = ApiField(alias='retireCertificates')

class KeyGenerationTargetsInformation(TargetsInformation):
    ...

class MonthlyPatternInformation(SchedulerPatternInformation):
    daysOfMonth: List[Literal[
        'D1', 'D10', 'D11', 'D12', 'D13', 'D14', 'D15', 'D16', 'D17', 'D18', 'D19', 'D2', 'D20', 'D21', 'D22', 'D23', 'D24', 'D25',
        'D26', 'D27', 'D28', 'D29', 'D3', 'D30', 'D31', 'D4', 'D5', 'D6', 'D7', 'D8', 'D9', 'L', 'LW']] = ApiField(
        alias='daysOfMonth',
        default_factory=list
    )
    recurrenceTime: datetime = ApiField(alias='recurrenceTime')

class SmartDiscoveryExternalTargetsInformation(TargetsInformation):
    ports: List[str] = ApiField(alias='ports', default_factory=list)
    retireCertificates: bool = ApiField(alias='retireCertificates')

BasicDiscoveryTargetsInformation.update_forward_refs()
CronPatternInformation.update_forward_refs()
EnhancedDiscoveryTargetsInformation.update_forward_refs()
EnvironmentInformation.update_forward_refs()
EnvironmentResponse.update_forward_refs()
ErrorInformation.update_forward_refs()
ErrorResponse.update_forward_refs()
ExternalScanTargetsInformation.update_forward_refs()
IntegrationServiceCreationRequest.update_forward_refs()
IntegrationServiceDetailsResponse.update_forward_refs()
IntegrationServiceInformation.update_forward_refs()
IntegrationServiceInitiateScanRequest.update_forward_refs()
IntegrationServiceUpdateRequest.update_forward_refs()
IntegrationsServicesAggregatesResponse.update_forward_refs()
InternetDiscoveryTargetsInformation.update_forward_refs()
KeyGenerationTargetsInformation.update_forward_refs()
MonthlyPatternInformation.update_forward_refs()
SchedulerPatternInformation.update_forward_refs()
SmartDiscoveryExternalTargetsInformation.update_forward_refs()
TargetsInformation.update_forward_refs()
WeeklyPatternInformation.update_forward_refs()



from __future__ import annotations

from pyvenafi.tpp.api.api_base import (
    ApiField,
    generate_output,
    WebSdkEndpoint,
    WebSdkOutputModel,
)
from pyvenafi.tpp.api.websdk.models import processing_engines

class _ProcessingEngines(WebSdkEndpoint):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/ProcessingEngines')
        self.Engine = self._Engine(api_obj=api_obj, url=f'{self._url}/Engine')
        self.Folder = self._Folder(api_obj=api_obj, url=f'{self._url}/Folder')

    def get(self):
        class Output(WebSdkOutputModel):
            engines: list[processing_engines.Engine] = ApiField(alias='Engines', default_factory=list)

        return generate_output(output_cls=Output, response=self._get())

    class _Engine(WebSdkEndpoint):
        def Guid(self, guid: str):
            return self._Guid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

        class _Guid(WebSdkEndpoint):
            def get(self):
                class Output(WebSdkOutputModel):
                    links: processing_engines.Link = ApiField(alias='_links')
                    folders: list[processing_engines.Folder] = ApiField(alias='Folders', default_factory=list)

                return generate_output(output_cls=Output, response=self._get())

            def post(self, folder_guids: list[str]):
                body = {
                    'FolderGuids': folder_guids
                }

                class Output(WebSdkOutputModel):
                    added_count: int = ApiField(alias='AddedCount')
                    errors: list[str] = ApiField(alias='Errors', default_factory=list)

                return generate_output(output_cls=Output, response=self._post(data=body))

    class _Folder(WebSdkEndpoint):
        def Guid(self, guid: str):
            return self._Guid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

        class _Guid(WebSdkEndpoint):
            def delete(self):
                return generate_output(output_cls=WebSdkOutputModel, response=self._delete())

            def get(self):
                class Output(WebSdkOutputModel):
                    links: processing_engines.Link = ApiField(alias='_links')
                    engines: list[processing_engines.Engine] = ApiField(alias='Engines', default_factory=list)

                return generate_output(output_cls=Output, response=self._get())

            def put(self, engine_guids: list[str]):
                body = {
                    'EngineGuids': engine_guids
                }

                return generate_output(output_cls=WebSdkOutputModel, response=self._put(data=body))

            def EngineGuid(self, guid):
                return self._EngineGuid(api_obj=self._api_obj, url=f'{self._url}/{guid}')

            class _EngineGuid(WebSdkEndpoint):
                def delete(self):
                    return generate_output(output_cls=WebSdkOutputModel, response=self._delete())



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.flow_action_approver_base import FlowActionApproverBaseAttributes

class FlowActionConfigReadApproversAttributes(FlowActionApproverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Flow Action Config Read Approvers"
    attribute_name = Attribute('Attribute Name', min_version='21.4')
    object_dn = Attribute('Object DN', min_version='21.4')
    policy_read = Attribute('Policy Read', min_version='21.4')
    use_administrators_as_fallback_approvers = Attribute('Use Administrators as Fallback Approvers', min_version='21.4')



from __future__ import annotations

class ConfigAttributeDataTypes:
    data_types = {
        0 : 'None',
        1 : 'Base64',
        2 : 'Boolean',
        3 : 'CaseExactString',
        4 : 'CaseIgnoreList',
        5 : 'CaseIgnoreString, String',
        6 : 'Counter',
        7 : 'DistinguishedName',
        8 : 'EMailAddress',
        9 : 'FaxNumber',
        10: 'Hold',
        11: 'Integer',
        12: 'Interval',
        13: 'IPNetworkAddress',
        14: 'NetworkAddress',
        15: 'NumericString',
        16: 'OctetList',
        17: 'OctetString',
        18: 'Path',
        19: 'PostalAddress',
        20: 'PrintableString',
        21: 'SchemaClass',
        22: 'Stream',
        23: 'TelephoneNumber',
        24: 'Time',
    }



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.driver_base import DriverBaseAttributes

class AgentDriverBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
    __config_class__ = "Agent Driver Base"
    address = Attribute('Address')
    agent_discovery_redirection_rule = Attribute('Agent Discovery Redirection Rule')
    credential = Attribute('Credential')
    expiration = Attribute('Expiration')
    extended_host_details = Attribute('Extended Host Details')
    interval = Attribute('Interval')
    log_threshold = Attribute('Log Threshold')
    managed_by_discovery_dn = Attribute('Managed By Discovery DN')
    module_disabled = Attribute('Module Disabled')
    port = Attribute('Port')
    version = Attribute('Version')



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.flow_action_config_read_approvers import FlowActionConfigReadApproversAttributes

class FlowActionTeamOwnerApprovalAttributes(FlowActionConfigReadApproversAttributes, metaclass=IterableMeta):
    __config_class__ = "Flow Action Team Owner Approval"



from __future__ import annotations
from pyvenafi.cloud.api.api_base import (
    ApiField,
    ObjectModel,
)
from datetime import datetime
from typing import (
    Any,
    Dict,
    List,
    Literal,
)
from uuid import UUID

AnyValue = Any
ApiKeyStatus = Literal[
    "PENDING_ACTIVATION",
    "ACTIVE",
    "INACTIVE",
    "PENDING_ROTATION",
    "ROTATED",
    "EXPIRED"
]
ApiVersion = Literal[
    "ALL",
    "V1"
]
Capability = Literal[
    "DEFAULT",
    "ISSUANCE",
    "VALIDATION",
    "ENTERPRISE",
    "PROVISIONING",
    "SINGLE_SIGN_ON",
    "DISTRIBUTED_ISSUER",
    "CLOUD_PROVIDERS",
    "CREDENTIAL_MANAGER",
    "KUBERNETES_DISCOVERY",
    "KUBERNETES_ENTERPRISE_COMPONENTS",
    "KUBERNETES_ENTERPRISE_COMPONENTS_CERT_MANAGER",
    "KUBERNETES_ENTERPRISE_COMPONENTS_VEI",
    "KUBERNETES_ENTERPRISE_COMPONENTS_APE",
    "KUBERNETES_INTEGRATION_TLSPDC"
]
CompanyType = Literal[
    "INTERNAL",
    "CA",
    "TPP_CUSTOMER",
    "REGULAR",
    "OTHER"
]
DayOfWeek = Literal[
    "MONDAY",
    "TUESDAY",
    "WEDNESDAY",
    "THURSDAY",
    "FRIDAY",
    "SATURDAY",
    "SUNDAY"
]
NotificationType = Literal[
    "CERTIFICATE_EXPIRATION",
    "APPLICATION_CERTIFICATE_EXPIRATION",
    "CERTIFICATE_STATUS_DIGEST"
]
Operator = Literal[
    "EQUALS",
    "NOT_EQUALS",
    "CONTAINS",
    "NOT_CONTAINS",
    "STARTS_WITH",
    "ENDS_WITH"
]
ProductEntitlement = Literal[
    "ANY",
    "MIRA",
    "DEVOPS",
    "OUTAGE_DETECTION",
    "CODESIGN"
]
RecurrenceType = Literal['DAY_OF_WEEK']
ReferralPartner = Literal[
    "globalsign-hvca",
    "digicert-certcentral"
]
Role = Literal[
    "SECURITY_ADMIN",
    "DEVOPS_LEAD",
    "DEVOPS_USER",
    "OUTAGEDETECTION_ADMIN",
    "RESOURCE_OWNER",
    "PKI_ADMIN",
    "GUEST",
    "PLATFORM_ADMIN"
]
SSOStatus = Literal[
    "ACTIVE",
    "INACTIVE"
]
SystemRole = Literal[
    "SYSTEM_ADMIN",
    "CONDOR_METRICS"
]
TeamRole = Literal[
    "SYSTEM_ADMIN",
    "PKI_ADMIN",
    "RESOURCE_OWNER",
    "GUEST",
    "PLATFORM_ADMIN"
]
UserAccountType = Literal[
    "WEB_UI",
    "API"
]
UserPreferenceName = Literal[
    "DASHBOARD_PERSONA",
    "NEWS_BANNER_SHOW",
    "RECENTLY_FOUND_CERTIFICATES_INTERVAL",
    "SELECTED_USER_ENTITLEMENT",
    "DEFAULT_PAGE_VIEW_CA_ACCOUNTS",
    "DEFAULT_PAGE_VIEW_APPLICATIONS",
    "DEFAULT_PAGE_VIEW_USERS",
    "DEFAULT_PAGE_VIEW_TEAMS",
    "DEFAULT_PAGE_VIEW_ORGANIZATIONAL_UNITS",
    "DEFAULT_PAGE_VIEW_OD_CERTIFICATES",
    "SAVED_QUERIES_ACTIVITY_LOGGING",
    "PREFERENCE_1",
    "PREFERENCE_2",
    "PREFERENCE_3",
    "PREFERENCE_4",
    "PREFERENCE_5",
    "PREFERENCE_6",
    "PREFERENCE_7"
]
UserStatus = Literal[
    "PENDING_ACTIVATION",
    "ACTIVE",
    "INACTIVE"
]
UserType = Literal[
    "EXTERNAL",
    "INTERNAL"
]

class ApiKeyInformation(ObjectModel):
    apiKeyStatus: ApiKeyStatus = ApiField(alias='apiKeyStatus')
    apiVersion: ApiVersion = ApiField(alias='apiVersion')
    companyId: UUID = ApiField(alias='companyId')
    creationDate: datetime = ApiField(alias='creationDate')
    key: UUID = ApiField(alias='key')
    memberedTeams: List[UUID] = ApiField(alias='memberedTeams', default_factory=list)
    productRoles: Dict[str, List[Role]] = ApiField(alias='productRoles', default_factory=dict)
    systemRoles: List[SystemRole] = ApiField(alias='systemRoles', default_factory=list)
    userId: UUID = ApiField(alias='userId')
    username: str = ApiField(alias='username')
    validityEndDate: datetime = ApiField(alias='validityEndDate')
    validityStartDate: datetime = ApiField(alias='validityStartDate')

class ApiKeyNullResponse(ObjectModel):
    key: UUID = ApiField(alias='key')
    username: str = ApiField(alias='username')

class ApiKeyRequest(ObjectModel):
    apiVersion: ApiVersion = ApiField(alias='apiVersion')
    validityDays: int = ApiField(alias='validityDays')

class ApiKeyResponse(ObjectModel):
    apiKeys: List[ApiKeyInformation] = ApiField(alias='apiKeys', default_factory=list)

class CapabilityInformation(ObjectModel):
    isTrial: bool = ApiField(alias='isTrial')
    name: Capability = ApiField(alias='name')
    productExpiryDate: datetime = ApiField(alias='productExpiryDate')

class ChangePasswordRequest(ObjectModel):
    currentPassword: str = ApiField(alias='currentPassword')
    newPassword: str = ApiField(alias='newPassword')

class CompanyInformation(ObjectModel):
    active: bool = ApiField(alias='active')
    companyType: CompanyType = ApiField(alias='companyType')
    creationDate: datetime = ApiField(alias='creationDate')
    domains: List[str] = ApiField(alias='domains', default_factory=list)
    id: UUID = ApiField(alias='id')
    name: str = ApiField(alias='name')
    productEntitlements: List[ProductEntitlementInformation] = ApiField(
        alias='productEntitlements',
        default_factory=list
    )
    referralPartner: ReferralPartner = ApiField(alias='referralPartner')
    urlPrefix: str = ApiField(alias='urlPrefix')

class CompanyLoginConfigResponse(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    newConfigEnabled: bool = ApiField(alias='newConfigEnabled')
    ssoLogin: bool = ApiField(alias='ssoLogin')

class CreateTeamRequest(ObjectModel):
    members: List[UUID] = ApiField(alias='members', default_factory=list)
    name: str = ApiField(alias='name')
    owners: List[UUID] = ApiField(alias='owners', default_factory=list)
    role: TeamRole = ApiField(alias='role')
    userMatchingRules: List[UserMatchingRule] = ApiField(alias='userMatchingRules', default_factory=list)

class ErrorInformation(ObjectModel):
    args: List[AnyValue] = ApiField(alias='args', default_factory=list)
    code: int = ApiField(alias='code')
    message: str = ApiField(alias='message')

class ErrorResponse(ObjectModel):
    errors: List[ErrorInformation] = ApiField(alias='errors', default_factory=list)

class InvitationConfirmationRequest(ObjectModel):
    firstname: str = ApiField(alias='firstname')
    grecaptchaResponse: str = ApiField(alias='grecaptchaResponse')
    invitationId: UUID = ApiField(alias='invitationId')
    lastname: str = ApiField(alias='lastname')
    password: str = ApiField(alias='password')
    username: str = ApiField(alias='username')

class InvitationInformation(ObjectModel):
    id: UUID = ApiField(alias='id')
    productRoles: Dict[str, List[Role]] = ApiField(alias='productRoles', default_factory=dict)
    properties: Dict[str, str] = ApiField(alias='properties', default_factory=dict)

class InvitationRequest(ObjectModel):
    properties: Dict[str, str] = ApiField(alias='properties', default_factory=dict)
    roles: Dict[str, Role] = ApiField(alias='roles', default_factory=dict)
    teams: List[UUID] = ApiField(alias='teams', default_factory=list)

class InvitationResponse(ObjectModel):
    invitations: List[InvitationInformation] = ApiField(alias='invitations', default_factory=list)

class LocalLoginRequest(ObjectModel):
    localLoginDisabled: bool = ApiField(alias='localLoginDisabled')

class NotificationConfigurationInformation(ObjectModel):
    id: UUID = ApiField(alias='id')
    recipients: List[str] = ApiField(alias='recipients', default_factory=list)
    recurrencePattern: RecurrencePatternInformation = ApiField(alias='recurrencePattern')
    type: NotificationType = ApiField(alias='type')

class NotificationConfigurationRequest(ObjectModel):
    recipients: List[str] = ApiField(alias='recipients', default_factory=list)
    recurrencePattern: RecurrencePatternInformation = ApiField(alias='recurrencePattern')
    type: NotificationType = ApiField(alias='type')

class NotificationConfigurationResponse(ObjectModel):
    configurations: List[NotificationConfigurationInformation] = ApiField(alias='configurations', default_factory=list)

class ProductEntitlementInformation(ObjectModel):
    capabilities: List[CapabilityInformation] = ApiField(alias='capabilities', default_factory=list)
    label: ProductEntitlement = ApiField(alias='label')
    visibilityConstraintsInformation: VisibilityConstraintsInformation = ApiField(
        alias='visibilityConstraintsInformation'
    )

class RecurrencePatternInformation(ObjectModel):
    recurrenceType: RecurrenceType = ApiField(alias='recurrenceType')
    recurrenceValues: List[DayOfWeek] = ApiField(alias='recurrenceValues', default_factory=list)

class ResendActivationRequest(ObjectModel):
    email: str = ApiField(alias='email')

class ResendActivationResponse(ObjectModel):
    user: UserInformation = ApiField(alias='user')

class ResetPasswordRequest(ObjectModel):
    email: str = ApiField(alias='email')

class ResetPasswordResponse(ObjectModel):
    message: str = ApiField(alias='message')

class RolesRequest(ObjectModel):
    roles: Dict[str, Role] = ApiField(alias='roles', default_factory=dict)
    systemAdmin: bool = ApiField(alias='systemAdmin')

class SsoConfigurationInformation(ObjectModel):
    clientId: str = ApiField(alias='clientId')
    clientSecret: str = ApiField(alias='clientSecret')
    companyId: UUID = ApiField(alias='companyId')
    id: UUID = ApiField(alias='id')
    issuerUrl: str = ApiField(alias='issuerUrl')
    newConfigEnabled: bool = ApiField(alias='newConfigEnabled')
    scopes: List[str] = ApiField(alias='scopes', default_factory=list)
    skipUserInfo: bool = ApiField(alias='skipUserInfo')

class SsoConfigurationRequest(ObjectModel):
    clientId: str = ApiField(alias='clientId')
    clientSecret: str = ApiField(alias='clientSecret')
    issuerUrl: str = ApiField(alias='issuerUrl')
    newConfigEnabled: bool = ApiField(alias='newConfigEnabled')
    scopes: List[str] = ApiField(alias='scopes', default_factory=list)
    skipUserInfo: bool = ApiField(alias='skipUserInfo')

class SsoConfigurationResponse(ObjectModel):
    ssoConfigurations: List[SsoConfigurationInformation] = ApiField(alias='ssoConfigurations', default_factory=list)

class TeamInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    id: UUID = ApiField(alias='id')
    members: List[UUID] = ApiField(alias='members', default_factory=list)
    modificationDate: datetime = ApiField(alias='modificationDate')
    name: str = ApiField(alias='name')
    owners: List[UUID] = ApiField(alias='owners', default_factory=list)
    productRoles: Dict[str, List[Role]] = ApiField(alias='productRoles', default_factory=dict)
    role: TeamRole = ApiField(alias='role')
    systemRoles: List[SystemRole] = ApiField(alias='systemRoles', default_factory=list)
    userMatchingRules: List[UserMatchingRule] = ApiField(alias='userMatchingRules', default_factory=list)

class TeamMembersRequest(ObjectModel):
    members: List[UUID] = ApiField(alias='members', default_factory=list)

class TeamOwnersRequest(ObjectModel):
    owners: List[UUID] = ApiField(alias='owners', default_factory=list)

class TeamsResponse(ObjectModel):
    teams: List[TeamInformation] = ApiField(alias='teams', default_factory=list)

class UnsubscribeNotificationInformation(ObjectModel):
    message: str = ApiField(alias='message')

class UpdatePasswordRequest(ObjectModel):
    newPassword: str = ApiField(alias='newPassword')
    token: str = ApiField(alias='token')

class UpdatePasswordResponse(ObjectModel):
    message: str = ApiField(alias='message')
    userId: UUID = ApiField(alias='userId')

class UpdateTeamRequest(ObjectModel):
    name: str = ApiField(alias='name')
    role: TeamRole = ApiField(alias='role')
    userMatchingRules: List[UserMatchingRule] = ApiField(alias='userMatchingRules', default_factory=list)

class UserAccountRequest(ObjectModel):
    companyName: str = ApiField(alias='companyName')
    firstname: str = ApiField(alias='firstname')
    grecaptchaResponse: str = ApiField(alias='grecaptchaResponse')
    lastname: str = ApiField(alias='lastname')
    marketoAttributes: Dict[str, str] = ApiField(alias='marketoAttributes', default_factory=dict)
    password: str = ApiField(alias='password')
    urlPrefix: str = ApiField(alias='urlPrefix')
    userAccountType: UserAccountType = ApiField(alias='userAccountType')
    username: str = ApiField(alias='username')

class UserAccountResponse(ObjectModel):
    apiKey: ApiKeyInformation = ApiField(alias='apiKey')
    company: CompanyInformation = ApiField(alias='company')
    user: UserInformation = ApiField(alias='user')

class UserAccountTypeRequest(ObjectModel):
    accountType: UserAccountType = ApiField(alias='accountType')

class UserDisabledRequest(ObjectModel):
    disabled: bool = ApiField(alias='disabled')

class UserInformation(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    creationDate: datetime = ApiField(alias='creationDate')
    disabled: bool = ApiField(alias='disabled')
    emailAddress: str = ApiField(alias='emailAddress')
    firstLoginDate: datetime = ApiField(alias='firstLoginDate')
    firstname: str = ApiField(alias='firstname')
    hasPassword: bool = ApiField(alias='hasPassword')
    id: UUID = ApiField(alias='id')
    lastname: str = ApiField(alias='lastname')
    localLoginDisabled: bool = ApiField(alias='localLoginDisabled')
    memberedTeams: List[UUID] = ApiField(alias='memberedTeams', default_factory=list)
    ownedTeams: List[UUID] = ApiField(alias='ownedTeams', default_factory=list)
    productRoles: Dict[str, List[Role]] = ApiField(alias='productRoles', default_factory=dict)
    signupAttributes: Dict[str, str] = ApiField(alias='signupAttributes', default_factory=dict)
    ssoStatus: SSOStatus = ApiField(alias='ssoStatus')
    systemRoles: List[SystemRole] = ApiField(alias='systemRoles', default_factory=list)
    userAccountType: UserAccountType = ApiField(alias='userAccountType')
    userStatus: UserStatus = ApiField(alias='userStatus')
    userType: UserType = ApiField(alias='userType')
    username: str = ApiField(alias='username')

class UserLoginConfigResponse(ObjectModel):
    companyId: UUID = ApiField(alias='companyId')
    localLogin: bool = ApiField(alias='localLogin')
    ssoLogin: bool = ApiField(alias='ssoLogin')

class UserMatchingRule(ObjectModel):
    claimName: str = ApiField(alias='claimName')
    operator: Operator = ApiField(alias='operator')
    value: str = ApiField(alias='value')

class UserPreferenceInformation(ObjectModel):
    id: UUID = ApiField(alias='id')
    name: UserPreferenceName = ApiField(alias='name')
    value: AnyValue = ApiField(alias='value')

class UserPreferenceRequest(ObjectModel):
    name: UserPreferenceName = ApiField(alias='name')
    value: AnyValue = ApiField(alias='value')

class UserPreferencesResponse(ObjectModel):
    preferences: List[UserPreferenceInformation] = ApiField(alias='preferences', default_factory=list)

class UserResponse(ObjectModel):
    users: List[UserInformation] = ApiField(alias='users', default_factory=list)

class VisibilityConstraintsInformation(ObjectModel):
    fullAccessDurationDays: int = ApiField(alias='fullAccessDurationDays')
    limitedAccessNotificationDays: int = ApiField(alias='limitedAccessNotificationDays')
    limitedVisibilityCertCount: int = ApiField(alias='limitedVisibilityCertCount')
    limitedVisibilityCertInstallPerCertCount: int = ApiField(alias='limitedVisibilityCertInstallPerCertCount')

ApiKeyInformation.update_forward_refs()
ApiKeyNullResponse.update_forward_refs()
ApiKeyRequest.update_forward_refs()
ApiKeyResponse.update_forward_refs()
CapabilityInformation.update_forward_refs()
ChangePasswordRequest.update_forward_refs()
CompanyInformation.update_forward_refs()
CompanyLoginConfigResponse.update_forward_refs()
CreateTeamRequest.update_forward_refs()
ErrorInformation.update_forward_refs()
ErrorResponse.update_forward_refs()
InvitationConfirmationRequest.update_forward_refs()
InvitationInformation.update_forward_refs()
InvitationRequest.update_forward_refs()
InvitationResponse.update_forward_refs()
LocalLoginRequest.update_forward_refs()
NotificationConfigurationInformation.update_forward_refs()
NotificationConfigurationRequest.update_forward_refs()
NotificationConfigurationResponse.update_forward_refs()
ProductEntitlementInformation.update_forward_refs()
RecurrencePatternInformation.update_forward_refs()
ResendActivationRequest.update_forward_refs()
ResendActivationResponse.update_forward_refs()
ResetPasswordRequest.update_forward_refs()
ResetPasswordResponse.update_forward_refs()
RolesRequest.update_forward_refs()
SsoConfigurationInformation.update_forward_refs()
SsoConfigurationRequest.update_forward_refs()
SsoConfigurationResponse.update_forward_refs()
TeamInformation.update_forward_refs()
TeamMembersRequest.update_forward_refs()
TeamOwnersRequest.update_forward_refs()
TeamsResponse.update_forward_refs()
UnsubscribeNotificationInformation.update_forward_refs()
UpdatePasswordRequest.update_forward_refs()
UpdatePasswordResponse.update_forward_refs()
UpdateTeamRequest.update_forward_refs()
UserAccountRequest.update_forward_refs()
UserAccountResponse.update_forward_refs()
UserAccountTypeRequest.update_forward_refs()
UserDisabledRequest.update_forward_refs()
UserInformation.update_forward_refs()
UserLoginConfigResponse.update_forward_refs()
UserMatchingRule.update_forward_refs()
UserPreferenceInformation.update_forward_refs()
UserPreferenceRequest.update_forward_refs()
UserPreferencesResponse.update_forward_refs()
UserResponse.update_forward_refs()
VisibilityConstraintsInformation.update_forward_refs()



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import IterableMeta
from pyvenafi.tpp.attributes.agent_driver_base import AgentDriverBaseAttributes
from pyvenafi.tpp.attributes.agent_ssh_base import AgentSSHBaseAttributes
from pyvenafi.tpp.attributes.schedule_base import ScheduleBaseAttributes

class AgentSSHDriverAttributes(
    AgentDriverBaseAttributes,
    AgentSSHBaseAttributes,
    ScheduleBaseAttributes,
    metaclass=IterableMeta
):
    __config_class__ = "Agent SSH Driver"



from __future__ import annotations

from pyvenafi.tpp.attributes._helper import (
    IterableMeta,
    Attribute,
)
from pyvenafi.tpp.attributes.service_module import ServiceModuleAttributes

class CertificateRevocationAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
    __config_class__ = "Certificate Revocation"
    ca_issuer_monitor_disabled = Attribute('CA Issuer Monitor Disabled', min_version='21.4')
    ocsp_concurrent_connection_limit = Attribute('OCSP Concurrent Connection Limit', min_version='21.4')
    ocsp_concurrent_request_limit = Attribute('OCSP Concurrent Request Limit', min_version='21.4')
